"""Repository for constant values in Datadoc backend."""

from datadoc_model.all_optional.model import LanguageStringType
from datadoc_model.all_optional.model import LanguageStringTypeItem

VALIDATION_ERROR = "Validation error: "

DATE_VALIDATION_MESSAGE = f"{VALIDATION_ERROR}contains_data_from must be the same or earlier date than contains_data_until"

OBLIGATORY_METADATA_WARNING = "Obligatory metadata is missing: "

INCONSISTENCIES_MESSAGE = "Inconsistencies found between extracted and existing metadata! This usually means that the new dataset has a different structure and that the version number should be incremented.\nDetails:"

OBLIGATORY_DATASET_METADATA_IDENTIFIERS: list = [
    "assessment",
    "dataset_state",
    "dataset_status",
    "name",
    "description",
    "population_description",
    "version",
    "version_description",
    "subject_field",
    "spatial_coverage_description",
    "owner",
    "contains_data_from",
    "contains_data_until",
    "contains_personal_data",
]

OBLIGATORY_DATASET_METADATA_IDENTIFIERS_MULTILANGUAGE = [
    "name",
    "description",
    "population_description",
    "version_description",
    "spatial_coverage_description",
]

OBLIGATORY_VARIABLES_METADATA_IDENTIFIERS = [
    "name",
    "data_type",
    "variable_role",
    "is_personal_data",
    "unit_type",
    "population_description",
    "data_source",
    "temporality_type",
]

OBLIGATORY_VARIABLES_PSEUDONYMIZATION_IDENTIFIERS = [
    "encryption_algorithm",
    "encryption_key_reference",
]


OBLIGATORY_VARIABLES_METADATA_IDENTIFIERS_MULTILANGUAGE = [
    "name",
]

DEFAULT_SPATIAL_COVERAGE_DESCRIPTION = LanguageStringType(
    [
        LanguageStringTypeItem(
            languageCode="nb",
            languageText="Norge",
        ),
        LanguageStringTypeItem(
            languageCode="nn",
            languageText="Noreg",
        ),
        LanguageStringTypeItem(
            languageCode="en",
            languageText="Norway",
        ),
    ],
)

NUM_OBLIGATORY_DATASET_FIELDS = len(OBLIGATORY_DATASET_METADATA_IDENTIFIERS)

NUM_OBLIGATORY_VARIABLES_FIELDS = len(OBLIGATORY_VARIABLES_METADATA_IDENTIFIERS)

DATASET_FIELDS_FROM_EXISTING_METADATA = [
    "dataset_status",
    "name",
    "description",
    "population_description",
    "subject_field",
    "keyword",
    "spatial_coverage_description",
    "use_restrictions",
    "custom_type",
    "owner",
    "version_description",
]

METADATA_DOCUMENT_FILE_SUFFIX = "__DOC.json"

PAPIS_STABLE_IDENTIFIER_TYPE = "FREG_SNR"
PAPIS_ENCRYPTION_KEY_REFERENCE = "papis-common-key-1"
DAEAD_ENCRYPTION_KEY_REFERENCE = "ssb-common-key-1"
ENCRYPTION_PARAMETER_SNAPSHOT_DATE = "snapshotDate"
ENCRYPTION_PARAMETER_KEY_ID = "keyId"
ENCRYPTION_PARAMETER_STRATEGY = "strategy"
ENCRYPTION_PARAMETER_STRATEGY_SKIP = "skip"
GS_PREFIX = "gs://"



"""Client for working with Variable Definitions at Statistics Norway."""

from ._generated.vardef_client import models
from ._generated.vardef_client.exceptions import *  # noqa: F403
from ._utils.constants import DEFAULT_DATE
from ._utils.constants import GENERATED_CONTACT
from .exceptions import VardefClientError
from .exceptions import VardefFileError
from .exceptions import VariableNotFoundError
from .vardef import Vardef
from .variable_definition import VariableDefinition



from pathlib import Path

import pytest

from dapla_metadata.standards.name_validator import ValidationResult
from dapla_metadata.standards.standard_validators import check_naming_standard
from dapla_metadata.standards.standard_validators import generate_validation_report
from dapla_metadata.standards.utils.constants import FILE_DOES_NOT_EXIST
from dapla_metadata.standards.utils.constants import FILE_IGNORED
from dapla_metadata.standards.utils.constants import INVALID_SYMBOLS
from dapla_metadata.standards.utils.constants import MISSING_DATA_STATE
from dapla_metadata.standards.utils.constants import MISSING_DATASET_SHORT_NAME
from dapla_metadata.standards.utils.constants import MISSING_PERIOD
from dapla_metadata.standards.utils.constants import MISSING_SHORT_NAME
from dapla_metadata.standards.utils.constants import MISSING_VERSION
from dapla_metadata.standards.utils.constants import NAME_STANDARD_SUCCESS
from dapla_metadata.standards.utils.constants import NAME_STANDARD_VIOLATION
from dapla_metadata.standards.utils.constants import PATH_IGNORED

pytest_plugins = ("pytest_asyncio",)


@pytest.mark.parametrize(
    ("file_path"),
    [
        "buckets/data/sirkus/utdata/person_data_p2021_v2.parquet",
    ],
)
@pytest.mark.asyncio
async def test_non_existent_path(file_path: str):
    results = await check_naming_standard(file_path=file_path)
    assert len(results) == 1
    assert results[0].messages[0] == FILE_DOES_NOT_EXIST
    assert results[0].success


@pytest.mark.parametrize(
    ("file_path"),
    [
        "ssb-staging-dapla-felles-data-delt/datadoc/utdata/person_data_v1.parquet",
        "dataset/klargjorte_data/arbmark/resources/person_data_v1.parquet",
    ],
)
@pytest.mark.asyncio
async def test_missing_date_period(file_path, tmp_path):
    full_path = tmp_path / file_path
    full_path.parent.mkdir(parents=True, exist_ok=True)
    full_path.touch()
    assert (await check_naming_standard(file_path=full_path))[0].violations == [
        MISSING_PERIOD,
    ]


@pytest.mark.parametrize(
    ("file_path"),
    [
        "gs://ssb-staging-dapla-felles-data-delt/stat_reg/person_data_p2022_v1.parquet",
        "gs://ssb-staging-dapla-felles-data-delt/datadoc/person_data_p2021_v3.parquet",
        "buckets/produkt/test-2/person_testdata_p2021-12-31_p2021-12-31_v1.parquet",
    ],
)
@pytest.mark.asyncio
async def test_missing_data_state(file_path, tmp_path):
    full_path = tmp_path / file_path
    full_path.parent.mkdir(parents=True, exist_ok=True)
    full_path.touch()
    assert (
        MISSING_DATA_STATE
        in (await check_naming_standard(file_path=full_path))[0].violations
    )


@pytest.mark.parametrize(
    ("file_path"),
    [
        "buckets/stat/inndata/person_data_p2022_v1.parquet",
        "gs://ssb-staging-dapla-felles-data-delt/inndata/person_data_p2022_v1.parquet",
        "gs://ssb-staging-dapla-felles-data-delt/klargjorte-data/person_data_p2021_v3.parquet",
        "buckets/produkt/utdata/person_testdata_p2021-12-31_p2021-12-31_v1.parquet",
    ],
)
@pytest.mark.asyncio
async def test_missing_shortname(file_path, tmp_path):
    test_file = tmp_path / file_path
    test_file.parent.mkdir(parents=True, exist_ok=True)
    test_file.touch()
    assert (await check_naming_standard(file_path=test_file))[0].violations == [
        MISSING_SHORT_NAME,
    ]


@pytest.mark.parametrize(
    ("file_path"),
    [
        "buckets/ssb-dapla-example-data-produkt-prod/ledstill/temp/skjema_v1.parquet",
        "gs:/ssb-dapla-example-data-produkt-prod/ledstill/temp/skjema_v1.parquet",
        "buckets/ssb-dapla-example-data-produkt-prod/ledstill/inndata/temp/skjema_p2018_p2020_v2.parquet",
        "gs://ssb-dapla-example-data-produkt-prod/temp/ledstill/inndata/temp/skjema_p2018_p2020",
        "gs://ssb-dapla-example-data-produkt-prod/konfigurasjon/ledstill/inndata/skjema_p2018_p2020",
    ],
)
@pytest.mark.asyncio
async def test_ignored_paths(file_path, tmp_path):
    full_path = tmp_path / file_path
    full_path.parent.mkdir(parents=True, exist_ok=True)
    full_path.touch()
    assert (await check_naming_standard(file_path=full_path))[0].messages[
        0
    ] == PATH_IGNORED


@pytest.mark.parametrize(
    ("file_path"),
    [
        "buckets/ssb-dapla-example-data-produkt-prod/ledstill/utdata/persån_testdata_p2021-12-31_p2021-12-31_v1.parquet",
        "gs://ssb-dapla-example-data-prædukt-prod/ledstill/utdata/person_testdata_p2021-12-31_p2021-12-31_v1.parquet",
    ],
)
@pytest.mark.asyncio
async def test_invalid_symbols(file_path, tmp_path):
    full_path = tmp_path / file_path
    full_path.parent.mkdir(parents=True, exist_ok=True)
    full_path.touch()
    result = await check_naming_standard(file_path=full_path)
    assert any(v for v in result[0].violations for v in [INVALID_SYMBOLS])


@pytest.mark.parametrize(
    ("file_path"),
    [
        "gs://ssb-staging-dapla-felles-data-delt/datadoc/utdata/p2021_v2.parquet",
        "buckets/produkt/datadoc/brukertest/1/sykefratot/klargjorte_data/_p2021-12-31_p2021-12-31_v1.parquet",
    ],
)
@pytest.mark.asyncio
async def test_missing_dataset_shortname(file_path, tmp_path):
    full_path = tmp_path / file_path
    full_path.parent.mkdir(parents=True, exist_ok=True)
    full_path.touch()
    result = await check_naming_standard(file_path=full_path)
    assert any(v for v in result[0].violations for v in [MISSING_DATASET_SHORT_NAME])


@pytest.mark.parametrize(
    ("file_path"),
    [
        "gs://ssb-staging-dapla-felles-data-delt/datadoc/utdata/p2021_v2.parquet",
        "buckets/produkt/datadoc/brukertest/1/sykefratot/klargjorte_data/_p2021-12-31_p2021-12-31_v1.parquet",
    ],
)
@pytest.mark.asyncio
async def test_missing_dataset_shortname_as_dict(file_path, tmp_path):
    full_path = tmp_path / file_path
    full_path.parent.mkdir(parents=True, exist_ok=True)
    full_path.touch()
    result = await check_naming_standard(file_path=full_path)
    assert not result[0].to_dict()["success"]


@pytest.mark.parametrize(
    ("file_path", "violations"),
    [
        (
            "gs://ssb-dapla-example-data-produkt-prod/ledstill/inndata/park_p2021-12-31_p2021-12-31.parquet",
            [MISSING_VERSION],
        ),
        (
            "gs://ssb-dapla-example-data-produkt-prod/inndata/skjema_v1.parquet",
            [MISSING_SHORT_NAME, MISSING_PERIOD],
        ),
        (
            "gs://inndata/skjema_v1.parquet",
            [MISSING_SHORT_NAME, MISSING_PERIOD],
        ),
        (
            "buckets/klargjorte_data/_p2021-12-31_p2021-12-31_v1.parquet",
            [MISSING_SHORT_NAME, MISSING_DATASET_SHORT_NAME],
        ),
        (
            "gs://klargjorte_data/_p2021-12-31_p2021-12-31_v1.parquet",
            [MISSING_SHORT_NAME, MISSING_DATASET_SHORT_NAME],
        ),
    ],
)
@pytest.mark.asyncio
async def test_missing_multiple(file_path: str, violations: list, tmp_path):
    full_path = tmp_path / file_path
    full_path.parent.mkdir(parents=True, exist_ok=True)
    full_path.touch()
    result = await check_naming_standard(file_path=full_path)
    if isinstance(result[0], ValidationResult):
        assert any(v for v in result[0].violations for v in violations)


@pytest.mark.parametrize(
    ("file_path"),
    [
        "buckets/produkt/datadoc/utdata/person_data_p2021_v2.parquet",
        "produkt/datadoc/utdata/person_data_p2021_p2022_v2.parquet",
        "datadoc/utdata/undermappe/person_data_p2021_v2.parquet",
        "delt-data/dataset/klargjorte_data/arbmark/resources/person_data_p2021-12-31_p2021-12-31_v1.parquet",
        "stat/inndata/person_testdata_p2021-12-31_p2021-12-31_v1.parquet",
        "ssb-delt/stat/klargjorte-data/person_testdata_p2021-12-31_p2021-12-31_v1.parquet",
        "produkt-delt/datadoc/brukertest/1/sykefratot/klargjorte_data/person_testdata_p2021-12-31_p2021-12-31_v1.parquet",
    ],
)
@pytest.mark.asyncio
async def test_check_naming_standard_specific_file_path(
    file_path: str,
    tmp_path: Path,
):
    full_path = tmp_path / file_path
    full_path.parent.mkdir(parents=True, exist_ok=True)
    full_path.touch()
    results = await check_naming_standard(
        full_path,
    )
    assert len(results) == 1
    for r in results:
        assert r.success
        assert NAME_STANDARD_SUCCESS in r.messages


@pytest.mark.parametrize(
    ("file_path"),
    [
        "buckets/produkt/datadoc/utdata/person_data_p2021_v2.csv",
        "samfunns-produkt/datadoc/brukertest/1/sykefratot/klargjorte_data/person_testdata_p2021-12-31_p2021-12-31_v1.json",
    ],
)
@pytest.mark.asyncio
async def test_check_naming_standard_ignored_file_type(
    file_path: str,
    tmp_path: Path,
):
    full_path = tmp_path / file_path
    full_path.parent.mkdir(parents=True, exist_ok=True)
    full_path.touch()
    results = await check_naming_standard(
        full_path,
    )
    assert len(results) == 1
    for r in results:
        assert r.success
        assert FILE_IGNORED in r.messages


@pytest.mark.parametrize(
    (
        "short_name",
        "files",
        "bucket_name",
    ),
    [
        (
            "stat_reg",
            [
                ("person_data_p2022_v1.parquet", NAME_STANDARD_SUCCESS),
                ("bil_data_p2022_v1.parquet", NAME_STANDARD_SUCCESS),
                ("my%stuff.csv", FILE_IGNORED),
            ],
            "ssb-staging-dapla-felles-data-delt",
        ),
        (
            "temp_stuff",
            [
                ("_p2022_v1.parquet", MISSING_DATASET_SHORT_NAME),
                ("bil_data_v1.parquet", MISSING_PERIOD),
                ("my%stuff.csv", FILE_IGNORED),
            ],
            "ssb-staging-dapla-felles-data-delt",
        ),
    ],
)
@pytest.mark.asyncio
async def test_check_naming_directory_with_subdirectories(
    short_name: str,
    files: list[tuple[str, str]],
    bucket_name: str,
    tmp_path: Path,
):
    fake_bucket = tmp_path / bucket_name / short_name

    data_states = ["inndata", "utdata"]

    for state in data_states:
        state_path = fake_bucket / state
        state_path.mkdir(parents=True, exist_ok=True)
        for file in files:
            (state_path / file[0]).touch()

    results = await check_naming_standard(
        tmp_path / bucket_name,
    )
    assert len(results) == 6
    for file in files:
        result = next(r for r in results if str(file[0]) in r.file_path)
        assert file[0] in result.file_path
        if file[1] in [NAME_STANDARD_SUCCESS, FILE_IGNORED]:
            assert file[1] in result.messages
        else:
            assert NAME_STANDARD_VIOLATION in result.messages
            assert file[1] in result.violations


@pytest.mark.parametrize(
    ("file_path", "violations"),
    [
        (
            "gs://ssb-dapla-example-data-produkt-prod/inndata/skjema_p2021-12-31_p2021-12-31_v1.parquet",
            [MISSING_SHORT_NAME],
        ),
        (
            "gs://ssb-dapla-example-data-produkt-prod/inndata/mappe/skjema_p2021-12-31_p2021-12-31_v1.parquet",
            [MISSING_SHORT_NAME],
        ),
        (
            "gs://inndata/skjema_p2021-12-31_p2021-12-31_v1.parquet",
            [MISSING_SHORT_NAME],
        ),
        (
            "buckets/produkt/klargjorte_data/bil_p2021-12-31_p2021-12-31_v1.parquet",
            [MISSING_SHORT_NAME],
        ),
        (
            "gs://ssb-dapla-example-data-produkt-prod/ledstill/skjema_p2021-12-31_p2021-12-31_v1.parquet",
            [MISSING_DATA_STATE],
        ),
        (
            "buckets/produkt/ledstill/bil_p2021-12-31_p2021-12-31_v1.parquet",
            [MISSING_DATA_STATE],
        ),
        (
            "buckets/bil_p2021-12-31_p2021-12-31_v1.parquet",
            [MISSING_SHORT_NAME, MISSING_DATA_STATE],
        ),
    ],
)
@pytest.mark.asyncio
async def test_validate_short_name(file_path: str, violations: list, tmp_path):
    full_path = tmp_path / file_path
    full_path.parent.mkdir(parents=True, exist_ok=True)
    full_path.touch()

    result = await check_naming_standard(file_path=full_path)
    if isinstance(result, ValidationResult):
        assert result.violations == violations


@pytest.mark.asyncio
async def test_generate_naming_standard_report(tmp_path):
    file_paths = [
        "buckets/ssb-dapla-example-data-produkt-prod/ledstill/inndata/skjema_p1988_v1.parquet",
        "buckets/ssb-dapla-example-data-produkt-prod/inndata/skjema_v2.parquet",
        "buckets/ssb-dapla-example-data-produkt-prod/utdata/editert_v1.parquet",
        "buckets/ssb-dapla-example-data-produkt-prod/klargjorte_data/_p2021-12-31_p2021-12-31_v1.parquet",
        "buckets/ssb-dapla-example-data-produkt-prod/ledstill/klargjorte_data/park_p2021-12-31_p2021-12-31_v1.parquet",
    ]
    for file_path in file_paths:
        full_path = tmp_path / file_path
        full_path.parent.mkdir(parents=True, exist_ok=True)
        full_path.touch()

    results = await check_naming_standard(
        file_path=str(tmp_path / "buckets/ssb-dapla-example-data-produkt-prod"),
    )

    if isinstance(results, list):
        report = generate_validation_report(validation_results=results)
        assert report.num_failures == 3
        assert report.num_files_validated == 5
        assert report.num_success == 2


@pytest.mark.asyncio
async def test_generate_naming_standard_report_failure(tmp_path):
    file_paths = [
        "buckets/ssb-dapla-example-data-produkt-prod/ledstill/skjema_p1988.parquet",
        "buckets/ssb-dapla-example-data-produkt-prod/skjema_v2.parquet",
        "buckets/ssb-dapla-example-data-produkt-prod/utdata/editert_v1.parquet",
        "buckets/ssb-dapla-example-data-produkt-prod/klargjorte_data/_p2021-12-31_p2021-12-31.parquet",
        "buckets/ssb-dapla-example-data-produkt-prod/ledstill/park_v1.parquet",
    ]
    for file_path in file_paths:
        full_path = tmp_path / file_path
        full_path.parent.mkdir(parents=True, exist_ok=True)
        full_path.touch()

    results = await check_naming_standard(
        file_path=str(tmp_path / "buckets/ssb-dapla-example-data-produkt-prod"),
    )

    if isinstance(results, list):
        report = generate_validation_report(validation_results=results)
        assert report.num_failures == 5
        assert report.num_files_validated == 5
        assert report.num_success == 0



import asyncio
import logging
import re
from collections.abc import AsyncGenerator

from upath import UPath
from upath.types import ReadablePathLike

from dapla_metadata.datasets.dapla_dataset_path_info import DaplaDatasetPathInfo
from dapla_metadata.datasets.dataset_parser import SUPPORTED_DATASET_FILE_SUFFIXES
from dapla_metadata.standards.utils.constants import FILE_DOES_NOT_EXIST
from dapla_metadata.standards.utils.constants import FILE_IGNORED
from dapla_metadata.standards.utils.constants import IGNORED_FOLDERS
from dapla_metadata.standards.utils.constants import INVALID_SYMBOLS
from dapla_metadata.standards.utils.constants import MISSING_DATA_STATE
from dapla_metadata.standards.utils.constants import MISSING_DATASET_SHORT_NAME
from dapla_metadata.standards.utils.constants import MISSING_PERIOD
from dapla_metadata.standards.utils.constants import MISSING_SHORT_NAME
from dapla_metadata.standards.utils.constants import MISSING_VERSION
from dapla_metadata.standards.utils.constants import NAME_STANDARD_SUCCESS
from dapla_metadata.standards.utils.constants import NAME_STANDARD_VIOLATION
from dapla_metadata.standards.utils.constants import PATH_IGNORED
from dapla_metadata.standards.utils.constants import SSB_NAMING_STANDARD_REPORT
from dapla_metadata.standards.utils.constants import SSB_NAMING_STANDARD_REPORT_FILES
from dapla_metadata.standards.utils.constants import (
    SSB_NAMING_STANDARD_REPORT_RESULT_AVERAGE,
)
from dapla_metadata.standards.utils.constants import (
    SSB_NAMING_STANDARD_REPORT_RESULT_BEST,
)
from dapla_metadata.standards.utils.constants import (
    SSB_NAMING_STANDARD_REPORT_RESULT_GOOD,
)
from dapla_metadata.standards.utils.constants import (
    SSB_NAMING_STANDARD_REPORT_RESULT_LOW,
)
from dapla_metadata.standards.utils.constants import (
    SSB_NAMING_STANDARD_REPORT_RESULT_NO_SCORE,
)
from dapla_metadata.standards.utils.constants import SSB_NAMING_STANDARD_REPORT_SUCCESS
from dapla_metadata.standards.utils.constants import (
    SSB_NAMING_STANDARD_REPORT_SUCCESS_RATE,
)
from dapla_metadata.standards.utils.constants import (
    SSB_NAMING_STANDARD_REPORT_VIOLATIONS,
)

logger = logging.getLogger(__name__)


class ValidationResult:
    """Result object for name standard validation."""

    def __init__(
        self,
        success: bool,
        file_path: str,
    ) -> None:
        """Initialize the validatation result."""
        self.success = success
        self.file_path = file_path
        self.messages: list[str] = []
        self.violations: list[str] = []

    def add_message(self, message: str) -> None:
        """Add message to list."""
        if message not in self.messages:
            self.messages.append(message)

    def add_violation(self, violation: str) -> None:
        """Add violation to list."""
        if violation not in self.violations:
            self.violations.append(violation)
        if self.success:
            self.success = False

    def __repr__(self) -> str:
        """Representation for debugging."""
        return f"ValidationResult(success={self.success}, file_path={self.file_path}, messages={self.messages}, violations={self.violations})"

    def to_dict(self) -> dict:
        """Return result as a dictionary."""
        return {
            "success": self.success,
            "file_path": self.file_path,
            "messages": self.messages,
            "violations": self.violations,
        }


class NamingStandardReport:
    """Report object for name standard validation."""

    def __init__(self, validation_results: list[ValidationResult]) -> None:
        """Initialize the naming standard report."""
        self.validation_results = validation_results
        self.num_files_validated = len(validation_results)
        self.num_success = len(
            [result for result in validation_results if result.success is True],
        )
        self.num_failures = len(
            [result for result in validation_results if result.success is False],
        )

    def generate_report(self) -> str:
        """Format the report as a string."""
        return (
            f"{SSB_NAMING_STANDARD_REPORT}\n"
            f"=============================\n"
            f"{self.evaluate_result()}"
            f"{SSB_NAMING_STANDARD_REPORT_SUCCESS_RATE}: {self.success_rate():.2f}%\n"
            f"{SSB_NAMING_STANDARD_REPORT_FILES}: {self.num_files_validated}\n"
            f"{SSB_NAMING_STANDARD_REPORT_SUCCESS}: {self.num_success}\n"
            f"{SSB_NAMING_STANDARD_REPORT_VIOLATIONS}s: {self.num_failures}\n"
        )

    def success_rate(self) -> int | float | None:
        """Calculate the success rate as a percentage.

        Returns:
            int | float | None: The success rate as a percentage, or None if
            no files were validated.
        """
        if self.num_files_validated == 0:
            return None
        return self.num_success / self.num_files_validated * 100

    def evaluate_result(self) -> str:
        """Returns an appropriate message based on the success rate."""
        rate = self.success_rate()
        if rate is not None:
            if 95 <= rate <= 100:
                return SSB_NAMING_STANDARD_REPORT_RESULT_BEST
            if 70 < rate < 95:
                return SSB_NAMING_STANDARD_REPORT_RESULT_GOOD
            if 40 <= rate <= 70:
                return SSB_NAMING_STANDARD_REPORT_RESULT_AVERAGE
            if rate < 40:
                return SSB_NAMING_STANDARD_REPORT_RESULT_LOW
        return SSB_NAMING_STANDARD_REPORT_RESULT_NO_SCORE


def _has_invalid_symbols(path: ReadablePathLike) -> bool:
    """Return True if string contains illegal symbols.

    Examples:
        >>> _has_invalid_symbols("åregang-øre")
        True

        >>> _has_invalid_symbols("Azor89")
        False

        >>> _has_invalid_symbols("ssbÆ-dapla-example-data-produkt-prod/ledstill/oppdrag/skjema_p2018_p2020_v1")
        True

        >>> _has_invalid_symbols("ssb-dapla-example-data-produkt-prod/ledstill/oppdrag/skjema_p2018_p2020_v1")
        False

        >>> _has_invalid_symbols("ssb-dapla-example-data-produkt-prod/ledstill/inndata/skjema_p2018_p202_v1/aar=2018/data.parquet")
        False
    """
    # TODO @mmwinther: The = symbol is allowed to avoid failures on subdirectories of partioned parquet datasets.
    # DPMETA-824
    return bool(re.search(r"[^a-zA-Z0-9\./:_\-=]", str(path).strip()))


def _check_violations(
    file: UPath,
) -> list[str]:
    """Check for missing attributes and invalid symbols."""
    path_info = DaplaDatasetPathInfo(file)
    checks = {
        MISSING_SHORT_NAME: path_info.statistic_short_name,
        MISSING_DATA_STATE: path_info.dataset_state,
        MISSING_PERIOD: path_info.contains_data_from,
        MISSING_DATASET_SHORT_NAME: path_info.dataset_short_name,
        MISSING_VERSION: path_info.dataset_version,
        INVALID_SYMBOLS: not _has_invalid_symbols(file),
    }

    return [message for message, value in checks.items() if not value]


async def _validate_file(
    file: UPath,
    check_file_exists: bool = False,
) -> ValidationResult:
    """Check for naming standard violations.

    Returns:
        A ValidationResult object containing messages and violations
    """
    logger.info("Validating file: %s", file)
    if file.suffix not in SUPPORTED_DATASET_FILE_SUFFIXES:
        logger.info("Skipping validation on non-dataset file: %s", file)
        return await _ignored_file_type_result(file)

    result = ValidationResult(success=True, file_path=str(file))

    if check_file_exists and not file.exists():
        result.add_message(
            FILE_DOES_NOT_EXIST,
        )

    result.violations = await asyncio.get_running_loop().run_in_executor(
        None,
        lambda: _check_violations(file),
    )

    if result.violations:
        result.success = False
        result.add_message(NAME_STANDARD_VIOLATION)
    else:
        result.success = True
        result.add_message(
            NAME_STANDARD_SUCCESS,
        )
    return result


async def _ignored_folder_result(file: UPath) -> ValidationResult:
    r = ValidationResult(success=True, file_path=str(file))
    r.add_message(PATH_IGNORED)
    return r


async def _ignored_file_type_result(file: UPath) -> ValidationResult:
    r = ValidationResult(success=True, file_path=str(file))
    r.add_message(FILE_IGNORED)
    return r


async def validate_directory(
    path: ReadablePathLike,
) -> AsyncGenerator[AsyncGenerator | asyncio.Task]:
    """Validate a file or recursively validate all files in a directory."""
    path = UPath(path)
    if set(path.parts).intersection(IGNORED_FOLDERS):
        logger.info("File path ignored: %s", path)
        yield asyncio.create_task(_ignored_folder_result(path))
    elif path.suffix:
        yield asyncio.create_task(_validate_file(path, check_file_exists=True))
    else:
        for obj in await asyncio.get_running_loop().run_in_executor(
            None,
            lambda: path.glob("*"),
        ):
            if obj.suffix:
                yield asyncio.create_task(_validate_file(obj), name=obj.name)
            else:
                logger.debug("Recursing into: %s", obj)
                yield validate_directory(obj)



"""Internal Variable Definitions Administration API

## Introduction  Variable Definitions are centralized definitions of concrete variables which are typically present in multiple datasets. Variable Definitions support standardization of data and metadata and facilitate sharing and joining of data by clarifying when variables have an identical definition.  ## Maintenance of Variable Definitions This API allows for creation, maintenance and access of Variable Definitions.  ### Ownership Creation and maintenance of variables may only be performed by Statistics Norway employees representing a specific Dapla team, who are defined as the owners of a given Variable Definition. The team an owner represents must be specified when making a request through the `active_group` query parameter. All maintenance is to be performed by the owners, with no intervention from administrators.  ### Status All Variable Definitions have an associated status. The possible values for status are `DRAFT`, `PUBLISHED_INTERNAL` and `PUBLISHED_EXTERNAL`.  #### Draft When a Variable Definition is created it is assigned the status `DRAFT`. Under this status the Variable Definition is:  - Only visible to Statistics Norway employees. - Mutable (it may be changed directly without need for versioning). - Not suitable to refer to from other systems.  This status may be changed to `PUBLISHED_INTERNAL` or `PUBLISHED_EXTERNAL` with a direct update.  #### Published Internal Under this status the Variable Definition is:  - Only visible to Statistics Norway employees. - Immutable (all changes are versioned). - Suitable to refer to in internal systems for statistics production. - Not suitable to refer to for external use (for example in Statistikkbanken).  This status may be changed to `PUBLISHED_EXTERNAL` by creating a Patch version.  #### Published External Under this status the Variable Definition is:  - Visible to the general public. - Immutable (all changes are versioned). - Suitable to refer to from any system.  This status may not be changed as it would break immutability. If a Variable Definition is no longer relevant then its period of validity should be ended by specifying a `valid_until` date in a Patch version.  ### Immutability Variable Definitions are immutable. This means that any changes must be performed in a strict versioning system. Consumers can avoid being exposed to breaking changes by specifying a `date_of_validity` when they request a Variable Definition.  #### Patches Patches are for changes which do not affect the fundamental meaning of the Variable Definition.  #### Validity Periods Validity Periods are versions with a period defined by a `valid_from` date and optionally a `valid_until` date. If the fundamental meaning of a Variable Definition is to be changed, it should be done by creating a new Validity Period.

The version of the OpenAPI document: 0.1
Contact: metadata@ssb.no
Generated by OpenAPI Generator (https://openapi-generator.tech)

Do not edit the class manually.
"""

from __future__ import annotations

import json
import pprint
from typing import Any

from pydantic import BaseModel
from pydantic import ConfigDict
from pydantic import ValidationError
from pydantic import field_validator
from typing_extensions import Self

from ..models.complete_response import CompleteResponse
from ..models.vardok_id_response import VardokIdResponse

GETVARDOKVARDEFMAPPINGBYID200RESPONSE_ONE_OF_SCHEMAS = [
    "CompleteResponse",
    "VardokIdResponse",
]


class GetVardokVardefMappingById200Response(BaseModel):
    """GetVardokVardefMappingById200Response"""

    # data type: CompleteResponse
    oneof_schema_1_validator: CompleteResponse | None = None
    # data type: VardokIdResponse
    oneof_schema_2_validator: VardokIdResponse | None = None
    actual_instance: CompleteResponse | VardokIdResponse | None = None
    one_of_schemas: set[str] = {"CompleteResponse", "VardokIdResponse"}

    model_config = ConfigDict(
        validate_assignment=True,
        protected_namespaces=(),
    )

    def __init__(self, *args, **kwargs) -> None:
        if args:
            if len(args) > 1:
                raise ValueError(
                    "If a position argument is used, only 1 is allowed to set `actual_instance`"
                )
            if kwargs:
                raise ValueError(
                    "If a position argument is used, keyword arguments cannot be used."
                )
            super().__init__(actual_instance=args[0])
        else:
            super().__init__(**kwargs)

    @field_validator("actual_instance")
    def actual_instance_must_validate_oneof(cls, v):
        instance = GetVardokVardefMappingById200Response.model_construct()
        error_messages = []
        match = 0
        # validate data type: CompleteResponse
        if not isinstance(v, CompleteResponse):
            error_messages.append(
                f"Error! Input type `{type(v)}` is not `CompleteResponse`"
            )
        else:
            match += 1
        # validate data type: VardokIdResponse
        if not isinstance(v, VardokIdResponse):
            error_messages.append(
                f"Error! Input type `{type(v)}` is not `VardokIdResponse`"
            )
        else:
            match += 1
        if match > 1:
            # more than 1 match
            raise ValueError(
                "Multiple matches found when setting `actual_instance` in GetVardokVardefMappingById200Response with oneOf schemas: CompleteResponse, VardokIdResponse. Details: "
                + ", ".join(error_messages)
            )
        if match == 0:
            # no match
            raise ValueError(
                "No match found when setting `actual_instance` in GetVardokVardefMappingById200Response with oneOf schemas: CompleteResponse, VardokIdResponse. Details: "
                + ", ".join(error_messages)
            )
        return v

    @classmethod
    def from_dict(cls, obj: str | dict[str, Any]) -> Self:
        return cls.from_json(json.dumps(obj))

    @classmethod
    def from_json(cls, json_str: str) -> Self:
        """Returns the object represented by the json string"""
        instance = cls.model_construct()
        error_messages = []
        match = 0

        # deserialize data into CompleteResponse
        try:
            instance.actual_instance = CompleteResponse.from_json(json_str)
            match += 1
        except (ValidationError, ValueError) as e:
            error_messages.append(str(e))
        # deserialize data into VardokIdResponse
        try:
            instance.actual_instance = VardokIdResponse.from_json(json_str)
            match += 1
        except (ValidationError, ValueError) as e:
            error_messages.append(str(e))

        if match > 1:
            # more than 1 match
            raise ValueError(
                "Multiple matches found when deserializing the JSON string into GetVardokVardefMappingById200Response with oneOf schemas: CompleteResponse, VardokIdResponse. Details: "
                + ", ".join(error_messages)
            )
        if match == 0:
            # no match
            raise ValueError(
                "No match found when deserializing the JSON string into GetVardokVardefMappingById200Response with oneOf schemas: CompleteResponse, VardokIdResponse. Details: "
                + ", ".join(error_messages)
            )
        return instance

    def to_json(self) -> str:
        """Returns the JSON representation of the actual instance"""
        if self.actual_instance is None:
            return "null"

        if hasattr(self.actual_instance, "to_json") and callable(
            self.actual_instance.to_json
        ):
            return self.actual_instance.to_json()
        return json.dumps(self.actual_instance)

    def to_dict(self) -> dict[str, Any] | CompleteResponse | VardokIdResponse | None:
        """Returns the dict representation of the actual instance"""
        if self.actual_instance is None:
            return None

        if hasattr(self.actual_instance, "to_dict") and callable(
            self.actual_instance.to_dict
        ):
            return self.actual_instance.to_dict()
        # primitive type
        return self.actual_instance

    def to_str(self) -> str:
        """Returns the string representation of the actual instance"""
        return pprint.pformat(self.model_dump())



from datetime import datetime
from typing import Any

import arrow

DOCUMENT_VERSION_KEY = "document_version"
DATADOC_KEY = "datadoc"
DATASET_KEY = "dataset"
VARIABLES_KEY = "variables"
PSEUDONYMIZATION_KEY = "pseudonymization"


class UnknownModelVersionError(Exception):
    """Exception raised for unknown model versions.

    This error is thrown when an unrecognized model version is encountered.
    """

    def __init__(
        self,
        supplied_version: str,
        *args: tuple[Any, ...],
    ) -> None:
        """Initialize the exception with the supplied version.

        Args:
            supplied_version: The version of the model that was not recognized.
            *args: Additional arguments for the Exception base class.
        """
        super().__init__(args)
        self.supplied_version = supplied_version

    def __str__(self) -> str:
        """Return string representation."""
        return f"Document Version ({self.supplied_version}) of discovered file is not supported"


def _convert_language_string_type(supplied_value: dict) -> list[dict[str, str]]:
    """Convert a dictionary of language-specific strings to a list of dictionaries.

    This function takes a dictionary with language codes as keys and
    corresponding language-specific strings as values, and converts it to a list
    of dictionaries with 'languageCode' and 'languageText' keys.

    Args:
        supplied_value: A dictionary containing language codes as keys and
            language strings as values.

    Returns:
        A list of dictionaries, each containing 'languageCode' and 'languageText'
        keys, representing the converted language strings.
    """
    return [
        {
            "languageCode": "en",
            "languageText": supplied_value["en"],
        },
        {
            "languageCode": "nn",
            "languageText": supplied_value["nn"],
        },
        {
            "languageCode": "nb",
            "languageText": supplied_value["nb"],
        },
    ]


def find_and_update_language_strings(supplied_metadata: dict | None) -> dict | None:
    """Find and update language-specific strings in the supplied metadata.

    This function iterates through the supplied metadata dictionary.
    For each key-value pair, if the value is a dictionary containing "en"
    it is passed to the `_convert_language_string_type` function to potentially
    update its format.

    Args:
        supplied_metadata: A metadata dictionary where values may include nested
            dictionaries with language-specific strings.

    Returns:
        The updated metadata dictionary. If the supplied metadata is not a
        dictionary, it returns `None`.
    """
    if isinstance(supplied_metadata, dict):
        for key, value in supplied_metadata.items():
            if isinstance(value, dict) and "en" in value:
                supplied_metadata[key] = _convert_language_string_type(value)
        return supplied_metadata
    return None


def remove_element_from_model(
    supplied_metadata: dict[str, Any],
    element_to_remove: str,
) -> None:
    """Remove an element from the supplied metadata dictionary.

    This function deletes a specified element from the supplied metadata dictionary
    if it exists.

    Args:
        supplied_metadata: The metadata dictionary from which the element will be
            removed.
        element_to_remove: The key of the element to be removed from the metadata
            dictionary.
    """
    supplied_metadata.pop(element_to_remove, None)


def cast_to_date_type(value_to_update: str | None) -> str | None:
    """Convert a string to a date string in ISO format.

    This function takes a string representing a date and converts it to a
    date string in ISO format. If the input is `None`, it returns `None` without
    modification.

    Args:
        value_to_update: A string representing a date or `None`.

    Returns:
        The date string in ISO format if the input was a valid date string, or
        `None` if the input was `None`.
    """
    if value_to_update is None:
        return value_to_update

    return str(
        arrow.get(
            value_to_update,
        ).date(),
    )


def convert_is_personal_data(supplied_metadata: dict[str, Any]) -> None:
    """Convert 'is_personal_data' values in the supplied metadata to boolean.

    Iterates over variables in the supplied metadata and updates the
    'is_personal_data' field:
      - Sets it to True for NON_PSEUDONYMISED_ENCRYPTED_PERSONAL_DATA and PSEUDONYMISED_ENCRYPTED_PERSONAL_DATA.
      - Sets it to False for NOT_PERSONAL_DATA.

    Args:
        supplied_metadata: The metadata dictionary to be updated.
    """
    for variable in supplied_metadata[DATADOC_KEY][VARIABLES_KEY]:
        value = variable["is_personal_data"]
        if value in (
            "NON_PSEUDONYMISED_ENCRYPTED_PERSONAL_DATA",
            "PSEUDONYMISED_ENCRYPTED_PERSONAL_DATA",
        ):
            variable["is_personal_data"] = True
        elif value == "NOT_PERSONAL_DATA":
            variable["is_personal_data"] = False


def copy_pseudonymization_metadata(supplied_metadata: dict[str, Any]) -> None:
    """Copies pseudonymization metadata from the old pseudonymization section into the corresponding variable.

    For each variable in `supplied_metadata[DATADOC_KEY][VARIABLES_KEY]` that has a matching
    `short_name` in `supplied_metadata[PSEUDONYMIZATION_KEY]["pseudo_variables"]`, this
    function copies the following fields into the variable's 'pseudonymization' dictionary:

        - stable_identifier_type
        - stable_identifier_version
        - encryption_algorithm
        - encryption_key_reference
        - encryption_algorithm_parameters

    From the pseudo_dataset the value dataset_pseudo_time is copied to each variable as pseudonymization_time.

    Args:
        supplied_metadata: The metadata dictionary to be updated.
    """
    pseudo_vars = supplied_metadata.get(PSEUDONYMIZATION_KEY, {}).get(
        "pseudo_variables", []
    )
    pseudo_dataset = (
        supplied_metadata.get(PSEUDONYMIZATION_KEY, {}).get("pseudo_dataset") or {}
    )
    pseudo_time = pseudo_dataset.get("dataset_pseudo_time", None)
    datadoc_vars = supplied_metadata.get(DATADOC_KEY, {}).get(VARIABLES_KEY, [])
    pseudo_lookup = {var.get("short_name"): var for var in pseudo_vars}

    for variable in datadoc_vars:
        short_name = variable.get("short_name")
        if short_name in pseudo_lookup:
            pseudo_var = pseudo_lookup[short_name]
            variable[PSEUDONYMIZATION_KEY] = variable.get(
                PSEUDONYMIZATION_KEY, {}
            ).copy()

            for field in [
                "stable_identifier_type",
                "stable_identifier_version",
                "encryption_algorithm",
                "encryption_key_reference",
                "encryption_algorithm_parameters",
            ]:
                variable[PSEUDONYMIZATION_KEY][field] = pseudo_var[field]
            variable[PSEUDONYMIZATION_KEY]["pseudonymization_time"] = pseudo_time

        else:
            variable[PSEUDONYMIZATION_KEY] = None


def convert_datetime_to_date(date_value: str | None) -> str | None:
    """Convert ISO datetime string to date string, handling None and invalid values."""
    if not date_value or not isinstance(date_value, str):
        return date_value

    try:
        dt = datetime.fromisoformat(date_value.replace("Z", "+00:00"))
        return dt.date().isoformat()
    except ValueError:
        return date_value


def add_container(existing_metadata: dict) -> dict:
    """Add container for previous versions.

    Adds a container structure for previous versions of metadata.
    This function wraps the existing metadata in a new container structure
    that includes the 'document_version', 'datadoc', and 'pseudonymization'
    fields. The 'document_version' is set to "0.0.1" and 'pseudonymization'
    is set to None.

    Args:
        existing_metadata: The original metadata dictionary to be wrapped.

    Returns:
        A new dictionary containing the wrapped metadata with additional fields.
    """
    return {
        DOCUMENT_VERSION_KEY: "0.0.1",
        DATADOC_KEY: existing_metadata,
        PSEUDONYMIZATION_KEY: None,
    }


def is_metadata_in_container_structure(
    metadata: dict,
) -> bool:
    """Check if the metadata is in the container structure.

    At a certain point a metadata 'container' was introduced.
    The container provides a structure for different 'types' of metadata, such as
    'datadoc', 'pseudonymization' etc.
    This function determines if the given metadata dictionary follows this container
    structure by checking for the presence of the 'datadoc' field.

    Args:
        metadata: The metadata dictionary to check.

    Returns:
        True if the metadata is in the container structure (i.e., contains the
        'datadoc' field), False otherwise.
    """
    return DATADOC_KEY in metadata



"""Configuration management for dataset package."""

from __future__ import annotations

import logging
import os
from pprint import pformat

from dotenv import dotenv_values
from dotenv import load_dotenv
from upath import UPath

from dapla_metadata._shared.enums import DaplaEnvironment
from dapla_metadata._shared.enums import DaplaRegion
from dapla_metadata._shared.enums import DaplaService

logger = logging.getLogger(__name__)

DOT_ENV_FILE_PATH = UPath(__file__).parent.joinpath(".env")

JUPYTERHUB_USER = "JUPYTERHUB_USER"
DAPLA_REGION = "DAPLA_REGION"
DAPLA_ENVIRONMENT = "DAPLA_ENVIRONMENT"
DAPLA_SERVICE = "DAPLA_SERVICE"
DAPLA_GROUP_CONTEXT = "DAPLA_GROUP_CONTEXT"
OIDC_TOKEN = "OIDC_TOKEN"  # noqa: S105


DATADOC_STATISTICAL_SUBJECT_SOURCE_URL_DEFAULT = (
    "https://www.ssb.no/xp/_/service/mimir/subjectStructurStatistics"
)


env_loaded = False


def _load_dotenv_file() -> None:
    global env_loaded  # noqa: PLW0603
    if not env_loaded and DOT_ENV_FILE_PATH.exists():
        load_dotenv(str(DOT_ENV_FILE_PATH))
        env_loaded = True
        logger.info(
            "Loaded .env file with config keys: \n%s",
            pformat(list(dotenv_values(str(DOT_ENV_FILE_PATH)).keys())),
        )


def get_config_item(item: str, *, raising: bool = False) -> str | None:
    """Get a config item. Makes sure all access is logged.

    Args:
        item: The name of the environment variable to obtain.
        raising: `True` if an exception should be raised when the item isn't present.

    Returns:
        The set value or `None`

    Raises:
        OSError: Only if `raising` is True and the item is not found.
    """
    _load_dotenv_file()
    value = os.environ.get(item)
    if raising and not value:
        msg = f"Environment variable {item} not defined."
        raise OSError(msg)
    logger.debug("Config accessed. %s", item)
    return value


def get_statistical_subject_source_url() -> str | None:
    """Get the URL to the statistical subject source."""
    return (
        get_config_item("DATADOC_STATISTICAL_SUBJECT_SOURCE_URL")
        or DATADOC_STATISTICAL_SUBJECT_SOURCE_URL_DEFAULT
    )


def get_dapla_region() -> DaplaRegion | None:
    """Get the Dapla region we're running on."""
    if region := get_config_item(DAPLA_REGION):
        return DaplaRegion(region)

    return None


def get_dapla_environment() -> DaplaEnvironment | None:
    """Get the Dapla environment we're running on."""
    if env := get_config_item(DAPLA_ENVIRONMENT):
        return DaplaEnvironment(env)

    return None


def get_dapla_service() -> DaplaService | None:
    """Get the Dapla service we're running on."""
    if service := get_config_item(DAPLA_SERVICE):
        return DaplaService(service)

    return None


def get_oidc_token(*, raising: bool = False) -> str | None:
    """Get the JWT token from the environment."""
    return get_config_item(OIDC_TOKEN, raising=raising)


def get_dapla_group_context(*, raising: bool = False) -> str | None:
    """Get the group which the user has chosen to represent."""
    return get_config_item(DAPLA_GROUP_CONTEXT, raising=raising)



import pytest
from pytest_mock import MockType

from dapla_metadata._shared.config import DAPLA_ENVIRONMENT
from dapla_metadata._shared.enums import DaplaEnvironment
from dapla_metadata.variable_definitions._utils.config import VARDEF_HOST_PROD
from dapla_metadata.variable_definitions._utils.config import VARDEF_HOST_TEST
from dapla_metadata.variable_definitions._utils.config import WORKSPACE_DIR
from dapla_metadata.variable_definitions._utils.config import (
    get_vardef_client_configuration,
)
from dapla_metadata.variable_definitions._utils.config import get_vardef_host
from dapla_metadata.variable_definitions._utils.config import get_workspace_dir
from dapla_metadata.variable_definitions.vardef import Vardef


@pytest.mark.parametrize(
    "env",
    [DaplaEnvironment.PROD.value],
)
def test_vardef_host_prod(monkeypatch: pytest.MonkeyPatch, env: str):
    monkeypatch.setenv(DAPLA_ENVIRONMENT, env)
    assert get_vardef_host() == VARDEF_HOST_PROD


@pytest.mark.parametrize(
    "env",
    [
        DaplaEnvironment.TEST.value,
        DaplaEnvironment.DEV.value,
    ],
)
def test_vardef_host_test(monkeypatch: pytest.MonkeyPatch, env: str):
    monkeypatch.setenv(DAPLA_ENVIRONMENT, env)
    assert get_vardef_host() == VARDEF_HOST_TEST


def test_vardef_host_default(monkeypatch: pytest.MonkeyPatch):
    monkeypatch.delenv(DAPLA_ENVIRONMENT, raising=False)
    monkeypatch.delenv("VARDEF_HOST", raising=False)
    assert get_vardef_host() == "http://localhost:8080"


def test_vardef_host_specified(monkeypatch: pytest.MonkeyPatch):
    monkeypatch.delenv(DAPLA_ENVIRONMENT, raising=False)
    monkeypatch.setenv("VARDEF_HOST", "my-host")
    assert get_vardef_host() == "my-host"


def test_get_vardef_client_configuration(
    monkeypatch: pytest.MonkeyPatch, fake_labid_jwt: str
):
    monkeypatch.setenv(DAPLA_ENVIRONMENT, DaplaEnvironment.TEST.value)
    config = get_vardef_client_configuration()
    assert config.host == get_vardef_host()
    assert config.access_token == fake_labid_jwt


def test_workspace_dir_set(monkeypatch: pytest.MonkeyPatch):
    monkeypatch.setenv(WORKSPACE_DIR, "home/work")
    assert get_workspace_dir() == "home/work"


@pytest.mark.parametrize("number_of_calls", [0, 3, 10])
def test_access_token_refreshed_for_each_request(
    mock_auth_client_fetch_personal_token: MockType, number_of_calls: int
):
    for _ in range(number_of_calls):
        Vardef.list_variable_definitions()
    assert mock_auth_client_fetch_personal_token.call_count == number_of_calls
    if number_of_calls:
        mock_auth_client_fetch_personal_token.assert_called_with(
            scopes=["all_groups", "current_group"], audiences=["vardef"]
        )



import pytest

from dapla_metadata._shared.config import DAPLA_GROUP_CONTEXT
from dapla_metadata._shared.config import DAPLA_REGION
from dapla_metadata._shared.config import OIDC_TOKEN
from dapla_metadata._shared.enums import DaplaRegion
from dapla_metadata.dapla import user_info
from dapla_metadata.dapla.user_info import DaplaLabUserInfo
from dapla_metadata.dapla.user_info import UnknownUserInfo
from dapla_metadata.dapla.user_info import UserInfo


@pytest.mark.parametrize(
    ("environment_variable_name", "environment_variable_value", "expected_class"),
    [
        (DAPLA_REGION, DaplaRegion.DAPLA_LAB.value, DaplaLabUserInfo),
        (None, None, UnknownUserInfo),
    ],
)
def test_get_user_info_for_current_platform(
    monkeypatch: pytest.MonkeyPatch,
    environment_variable_name: str,
    environment_variable_value: str,
    expected_class: type[UserInfo],
):
    if environment_variable_name:
        monkeypatch.setenv(environment_variable_name, environment_variable_value)
    assert isinstance(user_info.get_user_info_for_current_platform(), expected_class)


def test_dapla_lab_user_info_short_email(
    fake_jwt: str,
    raw_jwt_payload: dict[str, object],
    monkeypatch: pytest.MonkeyPatch,
):
    monkeypatch.setenv(OIDC_TOKEN, fake_jwt)
    assert DaplaLabUserInfo().short_email == raw_jwt_payload["email"]


def test_dapla_lab_user_info_short_email_no_jwt_available(
    monkeypatch: pytest.MonkeyPatch,
):
    monkeypatch.delenv(OIDC_TOKEN, raising=False)
    assert DaplaLabUserInfo().short_email is None


# Parameter is indirectly passed to the fake_jwt fixture
@pytest.mark.parametrize(("raw_jwt_payload"), [{"no_email": "no_email_in_jwt"}])
def test_dapla_lab_user_info_short_email_no_email_in_jwt(
    fake_jwt: str,
    monkeypatch: pytest.MonkeyPatch,
):
    monkeypatch.setenv(OIDC_TOKEN, fake_jwt)
    assert DaplaLabUserInfo().short_email is None


@pytest.mark.parametrize(
    ("environment_variable_name", "environment_variable_value", "expected_team"),
    [
        (DAPLA_GROUP_CONTEXT, "dapla-metadata-developers", "dapla-metadata"),
        (
            DAPLA_GROUP_CONTEXT,
            "dapla-metadata-data-admins",
            "dapla-metadata",
        ),
        (DAPLA_GROUP_CONTEXT, "dapla-metadata-dev", "dapla-metadata"),
        (DAPLA_GROUP_CONTEXT, "dapla-metadata", "dapla"),
        (
            DAPLA_GROUP_CONTEXT,
            "dapla-metadata-not-a-real-group",
            "dapla-metadata-not-a-real",
        ),
    ],
)
def test_get_current_team_dapla_lab(
    monkeypatch: pytest.MonkeyPatch,
    environment_variable_name: str,
    environment_variable_value: str,
    expected_team: str,
):
    monkeypatch.setenv(DAPLA_REGION, DaplaRegion.DAPLA_LAB)
    if environment_variable_name:
        monkeypatch.setenv(environment_variable_name, environment_variable_value)
    assert user_info.get_user_info_for_current_platform().current_team == expected_team


@pytest.mark.parametrize(
    ("environment_variable_value"),
    [
        None,
        "",
    ],
)
def test_get_owner_errors_dapla_lab(
    monkeypatch: pytest.MonkeyPatch,
    environment_variable_value: str | None,
):
    monkeypatch.setenv(DAPLA_REGION, DaplaRegion.DAPLA_LAB)
    if environment_variable_value is None:
        monkeypatch.delenv(DAPLA_GROUP_CONTEXT, raising=False)
    else:
        monkeypatch.setenv(DAPLA_GROUP_CONTEXT, environment_variable_value)

    with pytest.raises(
        OSError,
        match="DAPLA_GROUP_CONTEXT environment variable not found",
    ):
        user_info.get_user_info_for_current_platform().current_team  # noqa: B018



import string
import sys
import typing as t

import jwt
import pytest
from faker import Faker

if sys.version_info < (3, 11):
    # dapla_auth_client only supports Python >=3.11
    # We must still support 3.10 for now, but variable definitions never runs on 3.10
    collect_ignore_glob = ["*variable_definitions*"]


@pytest.fixture
def raw_jwt_payload(faker: Faker) -> dict[str, t.Any]:
    user_name = "".join(faker.random_sample(elements=string.ascii_lowercase, length=3))
    email = f"{user_name}@ssb.no"
    first_name = faker.first_name()
    last_name = faker.last_name()
    return {
        "exp": faker.unix_time(),
        "iat": faker.unix_time(),
        "auth_time": faker.unix_time(),
        "jti": faker.uuid4(),
        "iss": faker.url(),
        "aud": [
            faker.word(),
            faker.uuid4(),
            "broker",
            "account",
        ],
        "sub": faker.uuid4(),
        "typ": "Bearer",
        "azp": "onyxia",
        "session_state": faker.uuid4(),
        "allowed-origins": ["*"],
        "realm_access": {
            "roles": [faker.word(), faker.word()],
        },
        "resource_access": {
            "broker": {"roles": [faker.word()]},
            "account": {
                "roles": [faker.word()],
            },
        },
        "scope": "openid email profile",
        "sid": faker.uuid4(),
        "email_verified": True,
        "name": f"{first_name} {last_name}",
        "short_username": f"ssb-{user_name}",
        "preferred_username": email,
        "given_name": first_name,
        "family_name": last_name,
        "email": email,
    }


@pytest.fixture
def fake_jwt(raw_jwt_payload: dict[str, t.Any]) -> str:
    return jwt.encode(raw_jwt_payload, "test secret", algorithm="HS256")


@pytest.fixture
def raw_labid_jwt_payload(faker: Faker) -> dict[str, t.Any]:
    return {
        "aud": ["vardef"],
        "dapla.group": "team-a-developers",
        "dapla.groups": [
            "team-a-data-admins",
            "team-a-developers",
            "team-b-developers",
        ],
        "exp": faker.unix_time(),
        "iat": faker.unix_time(),
        "iss": "https://labid.lab.dapla-external.ssb.no",
        "scope": "all_groups,current_group",
        "sub": "".join(faker.random_sample(elements=string.ascii_lowercase, length=3)),
    }


@pytest.fixture
def fake_labid_jwt(raw_labid_jwt_payload: dict[str, t.Any]) -> str:
    return jwt.encode(raw_labid_jwt_payload, "test secret", algorithm="HS256")



"""Constants used in validate ssb name standard."""

from dapla_metadata.datasets.dataset_parser import SUPPORTED_DATASET_FILE_SUFFIXES

SUCCESS = "Suksess"

NAME_STANDARD_SUCCESS = "Filene dine er i samsvar med SSB-navnestandarden"

NAME_STANDARD_VIOLATION = "Det er oppdaget brudd på SSB-navnestandard:"

MISSING_BUCKET_NAME = "Filnavn mangler bøttenavn ref: https://manual.dapla.ssb.no/statistikkere/navnestandard.html#obligatoriske-mapper"
MISSING_VERSION = "Filnavn mangler versjon, hvis ikke filen er nyeste versjon kan dette være brudd på navnestandarden ref: https://manual.dapla.ssb.no/statistikkere/navnestandard.html#versjonering-av-datasett"
MISSING_PERIOD = "Filnavn mangler gyldighetsperiode ref: https://manual.dapla.ssb.no/statistikkere/navnestandard.html#filnavn"
MISSING_SHORT_NAME = "Kortnavn for statistikk mangler ref: https://manual.dapla.ssb.no/statistikkere/navnestandard.html#obligatoriske-mapper"
MISSING_DATA_STATE = "Mappe for datatilstand mangler ref: https://manual.dapla.ssb.no/statistikkere/navnestandard.html#obligatoriske-mapper"
MISSING_DATASET_SHORT_NAME = "Filnavn mangler datasett kortnavn ref: https://manual.dapla.ssb.no/statistikkere/navnestandard.html#filnavn"

INVALID_SYMBOLS = "Filnavn inneholder ulovlige tegn ref: https://manual.dapla.ssb.no/statistikkere/navnestandard.html#filnavn"

PATH_IGNORED = "Ignorert, mappen er ikke underlagt krav til navnestandard."
FILE_IGNORED = f"Ignorert, kun datasett med {', '.join(SUPPORTED_DATASET_FILE_SUFFIXES.keys())} filendelser valideres foreløpig."

FILE_DOES_NOT_EXIST = "Filen eksisterer ikke. Validerer uansett."

BUCKET_NAME_UNKNOWN = "Kan ikke validere bøttenavn"

SSB_NAMING_STANDARD_REPORT = "SSB navnestandard rapport"
SSB_NAMING_STANDARD_REPORT_SUCCESS_RATE = "Suksess rate"
SSB_NAMING_STANDARD_REPORT_RESULT_BEST = "🚀 Fantastisk! 🎉\n"
SSB_NAMING_STANDARD_REPORT_RESULT_GOOD = (
    "✅ Bra jobba! Fortsatt litt rom for forbedring. 😊\n"
)
SSB_NAMING_STANDARD_REPORT_RESULT_AVERAGE = (
    "⚠️ Ikke verst! Men det er noen feil å fikse. 🔧\n"
)
SSB_NAMING_STANDARD_REPORT_RESULT_LOW = "❌ Mye å forbedre! Ta en grundig sjekk. 🛠️\n"
SSB_NAMING_STANDARD_REPORT_RESULT_NO_SCORE = "👀 Ingen filer ble validert\n"
SSB_NAMING_STANDARD_REPORT_FILES = "Antall filer validert"
SSB_NAMING_STANDARD_REPORT_SUCCESS = "Antall filer som følger SSB navnestandard"
SSB_NAMING_STANDARD_REPORT_VIOLATIONS = "Antall filer som bryter SSB navnestandard"

IGNORED_FOLDERS = [
    "temp",
    "oppdrag",
    "konfigurasjon",
    "logg",
    "tidsserier",
    "migrert",
]



"""Handle reading, updating and writing of metadata."""

from __future__ import annotations

import copy
import json
import logging
from concurrent.futures import ThreadPoolExecutor
from typing import TYPE_CHECKING
from typing import cast

import datadoc_model.all_optional.model as all_optional_model
import datadoc_model.required.model as required_model
from datadoc_model.all_optional.model import DataSetStatus
from upath import UPath

from dapla_metadata._shared import config
from dapla_metadata.dapla import user_info
from dapla_metadata.datasets._merge import DatasetConsistencyStatus
from dapla_metadata.datasets._merge import check_dataset_consistency
from dapla_metadata.datasets._merge import check_ready_to_merge
from dapla_metadata.datasets._merge import check_variables_consistency
from dapla_metadata.datasets._merge import merge_metadata
from dapla_metadata.datasets.compatibility import is_metadata_in_container_structure
from dapla_metadata.datasets.compatibility import upgrade_metadata
from dapla_metadata.datasets.dapla_dataset_path_info import DaplaDatasetPathInfo
from dapla_metadata.datasets.dataset_parser import DatasetParser
from dapla_metadata.datasets.model_validation import ValidateDatadocMetadata
from dapla_metadata.datasets.statistic_subject_mapping import StatisticSubjectMapping
from dapla_metadata.datasets.utility.constants import (
    DEFAULT_SPATIAL_COVERAGE_DESCRIPTION,
)
from dapla_metadata.datasets.utility.constants import METADATA_DOCUMENT_FILE_SUFFIX
from dapla_metadata.datasets.utility.constants import NUM_OBLIGATORY_DATASET_FIELDS
from dapla_metadata.datasets.utility.constants import NUM_OBLIGATORY_VARIABLES_FIELDS
from dapla_metadata.datasets.utility.urn import convert_uris_to_urns
from dapla_metadata.datasets.utility.urn import klass_urn_converter
from dapla_metadata.datasets.utility.urn import vardef_urn_converter
from dapla_metadata.datasets.utility.utils import OptionalDatadocMetadataType
from dapla_metadata.datasets.utility.utils import PseudonymizationType
from dapla_metadata.datasets.utility.utils import VariableListType
from dapla_metadata.datasets.utility.utils import VariableType
from dapla_metadata.datasets.utility.utils import calculate_percentage
from dapla_metadata.datasets.utility.utils import derive_assessment_from_state
from dapla_metadata.datasets.utility.utils import get_timestamp_now
from dapla_metadata.datasets.utility.utils import (
    num_obligatory_dataset_fields_completed,
)
from dapla_metadata.datasets.utility.utils import (
    num_obligatory_variables_fields_completed,
)
from dapla_metadata.datasets.utility.utils import set_dataset_owner
from dapla_metadata.datasets.utility.utils import set_default_values_dataset
from dapla_metadata.datasets.utility.utils import set_default_values_pseudonymization
from dapla_metadata.datasets.utility.utils import set_default_values_variables

if TYPE_CHECKING:
    from datetime import datetime

    from upath.types import ReadablePathLike

logger = logging.getLogger(__name__)


class Datadoc:
    """Handle reading, updating and writing of metadata.

    If a metadata document exists, it is this information that is loaded. Nothing
    is inferred from the dataset. If only a dataset path is supplied the metadata
    document path is build based on the dataset path.

    Example: /path/to/dataset.parquet -> /path/to/dataset__DOC.json

    Attributes:
        dataset_path: A file path to the path to where the dataset is stored.
        metadata_document_path: A path to a metadata document if it exists.
        statistic_subject_mapping: An instance of StatisticSubjectMapping.
    """

    def __init__(
        self,
        dataset_path: ReadablePathLike | None = None,
        metadata_document_path: ReadablePathLike | None = None,
        statistic_subject_mapping: StatisticSubjectMapping | None = None,
        errors_as_warnings: bool = False,
        validate_required_fields_on_existing_metadata: bool = False,
    ) -> None:
        """Initialize the Datadoc instance.

        If a dataset path is supplied, it attempts to locate and load the
        corresponding metadata document. If no dataset path is provided, the class
        is instantiated without loading any metadata.

        Args:
            dataset_path: The file path to the dataset. Defaults to None.
            metadata_document_path: The file path to the metadata document.
                Defaults to None.
            statistic_subject_mapping: An instance of StatisticSubjectMapping.
                Defaults to None
            errors_as_warnings: Disable raising exceptions if inconsistencies
                are found between existing and extracted metadata.
            validate_required_fields_on_existing_metadata: Use a Pydantic model
                which validates whether required fields are present when reading
                in an existing metadata file.
        """
        self._statistic_subject_mapping = statistic_subject_mapping
        self.errors_as_warnings = errors_as_warnings
        self.validate_required_fields_on_existing_metadata = (
            validate_required_fields_on_existing_metadata
        )
        self.metadata_document: UPath | None = None
        self.container: all_optional_model.MetadataContainer | None = None
        self.dataset_path: UPath | None = None
        self.dataset = all_optional_model.Dataset()
        self.variables: VariableListType = []
        self.variables_lookup: dict[str, VariableType] = {}
        self.explicitly_defined_metadata_document = False
        self.dataset_consistency_status: list[DatasetConsistencyStatus] = []
        if metadata_document_path:
            self.metadata_document = UPath(metadata_document_path)
            self.explicitly_defined_metadata_document = True
            if not self.metadata_document.exists():
                msg = f"Metadata document does not exist! Provided path: {self.metadata_document}"
                raise ValueError(
                    msg,
                )
        if dataset_path:
            self.dataset_path = UPath(dataset_path)
            if not metadata_document_path:
                self.metadata_document = self.build_metadata_document_path(
                    self.dataset_path,
                )
        if metadata_document_path or dataset_path:
            self._extract_metadata_from_files()

    def _extract_metadata_from_files(self) -> None:
        """Read metadata from an existing metadata document or create one.

        If a metadata document exists, it reads and extracts metadata from it.
        If no metadata document is found, it creates metadata from scratch by
        extracting information from the dataset file.

        This method ensures that:
        - Metadata is extracted from an existing document if available.
        - If metadata is not available, it is extracted from the dataset file.
        - The dataset ID is set if not already present.
        - Default values are set for variables, particularly the variable role on
            creation.
        - Default values for variables ID and 'is_personal_data' are set if the
            values are None.
        - The 'contains_personal_data' attribute is set to False if not specified.
        - A lookup dictionary for variables is created based on their short names.
        """
        extracted_metadata: all_optional_model.DatadocMetadata | None = None
        existing_metadata: OptionalDatadocMetadataType = None

        if self.metadata_document and self.metadata_document.exists():
            existing_metadata = self._extract_metadata_from_existing_document(
                self.metadata_document,
            )

        if (
            self.dataset_path is not None
            and self.dataset == all_optional_model.Dataset()
            and len(self.variables) == 0
        ):
            extracted_metadata = self._extract_metadata_from_dataset(self.dataset_path)

        if (
            self.dataset_path
            and self.metadata_document
            and extracted_metadata
            and existing_metadata
        ):
            self.dataset_consistency_status = check_dataset_consistency(
                self.dataset_path,
                self.metadata_document,
            )
            self.dataset_consistency_status.extend(
                check_variables_consistency(
                    extracted_metadata.variables or [],
                    existing_metadata.variables or [],
                )
            )

        if (
            self.dataset_path
            and self.explicitly_defined_metadata_document
            and self.metadata_document is not None
            and self.metadata_document.exists()
            and extracted_metadata is not None
            and existing_metadata is not None
        ):
            check_ready_to_merge(
                self.dataset_consistency_status,
                errors_as_warnings=self.errors_as_warnings,
            )
            merged_metadata = merge_metadata(
                extracted_metadata,
                existing_metadata,
            )
            # We need to override this so that the document gets saved to the correct
            # location, otherwise we would overwrite the existing document!
            self.metadata_document = self.build_metadata_document_path(
                self.dataset_path,
            )
            self._set_metadata(merged_metadata)
        else:
            self._set_metadata(existing_metadata or extracted_metadata)

    def _set_metadata(
        self,
        metadata: OptionalDatadocMetadataType,
    ) -> None:
        if not metadata or not (metadata.dataset and metadata.variables):
            msg = "Could not read metadata"
            raise ValueError(msg)
        self.dataset = cast("all_optional_model.Dataset", metadata.dataset)
        self.variables = metadata.variables

        set_default_values_variables(self.variables)
        set_default_values_dataset(cast("all_optional_model.Dataset", self.dataset))
        set_dataset_owner(self.dataset)
        convert_uris_to_urns(self.variables, "definition_uri", [vardef_urn_converter])
        convert_uris_to_urns(
            self.variables, "classification_uri", [klass_urn_converter]
        )
        self._create_variables_lookup()

    def _create_variables_lookup(self) -> None:
        self.variables_lookup = {
            v.short_name: v for v in self.variables if v.short_name
        }

    def _extract_metadata_from_existing_document(
        self,
        document: UPath,
    ) -> OptionalDatadocMetadataType:
        """Read metadata from an existing metadata document.

        If an existing metadata document is available, this method reads and
        loads the metadata from it. It validates and upgrades the metadata as
        necessary. If we have read in a file with an empty "datadoc" structure
        the process ends.
        A typical example causing a empty datadoc is a file produced from a
        pseudonymization process.

        Args:
            document: A path to the existing metadata document.

        Raises:
            json.JSONDecodeError: If the metadata document cannot be parsed.
            pydantic.ValidationError: If the data does not successfully validate.
        """
        metadata_model = (
            required_model
            if self.validate_required_fields_on_existing_metadata
            else all_optional_model
        )
        fresh_metadata = {}
        try:
            with document.open(mode="r", encoding="utf-8") as file:
                fresh_metadata = json.load(file)
            logger.info("Opened existing metadata file %s", document)
            fresh_metadata = upgrade_metadata(
                fresh_metadata,
            )
            if is_metadata_in_container_structure(fresh_metadata):
                self.container = metadata_model.MetadataContainer.model_validate_json(
                    json.dumps(fresh_metadata),
                )
                datadoc_metadata = fresh_metadata["datadoc"]
            else:
                datadoc_metadata = fresh_metadata
            if datadoc_metadata is None:
                return None
            return metadata_model.DatadocMetadata.model_validate_json(
                json.dumps(datadoc_metadata),
            )
        except json.JSONDecodeError:
            logger.warning(
                "Could not open existing metadata file %s. \
                    Falling back to collecting data from the dataset",
                document,
                exc_info=True,
            )
            return None

    def _extract_subject_field_from_path(
        self,
        dapla_dataset_path_info: DaplaDatasetPathInfo,
    ) -> str | None:
        """Extract the statistic short name from the dataset file path.

        Map the extracted statistic short name to its corresponding statistical
        subject.

        Args:
            dapla_dataset_path_info: The object representing the decomposed file
                path.

        Returns:
            The code for the statistical subject or None if we couldn't map to one.
        """
        if self._statistic_subject_mapping is None:
            with ThreadPoolExecutor(max_workers=12) as executor:
                return StatisticSubjectMapping(
                    executor,
                    config.get_statistical_subject_source_url(),
                ).get_secondary_subject(
                    dapla_dataset_path_info.statistic_short_name,
                )
        else:
            return self._statistic_subject_mapping.get_secondary_subject(
                dapla_dataset_path_info.statistic_short_name,
            )

    def _extract_metadata_from_dataset(
        self,
        dataset: UPath,
    ) -> all_optional_model.DatadocMetadata:
        """Obtain what metadata we can from the dataset itself.

        This makes it easier for the user by 'pre-filling' certain fields.
        Certain elements are dependent on the dataset being saved according
        to SSB's standard.

        Args:
            dataset: The path to the dataset file, which can be a local or
                cloud path.

        Side Effects:
            Updates the following instance attributes:
                - ds_schema: An instance of DatasetParser initialized for the
                    given dataset file.
                - dataset: An instance of model.Dataset with pre-filled metadata
                    fields.
                - variables: A list of fields extracted from the dataset schema.
        """
        dapla_dataset_path_info = DaplaDatasetPathInfo(dataset)
        metadata = all_optional_model.DatadocMetadata()

        metadata.dataset = all_optional_model.Dataset(
            short_name=dapla_dataset_path_info.dataset_short_name,
            dataset_state=dapla_dataset_path_info.dataset_state,
            dataset_status=DataSetStatus.DRAFT,
            assessment=(
                derive_assessment_from_state(
                    dapla_dataset_path_info.dataset_state,
                )
                if dapla_dataset_path_info.dataset_state is not None
                else None
            ),
            version=dapla_dataset_path_info.dataset_version,
            contains_data_from=dapla_dataset_path_info.contains_data_from,
            contains_data_until=dapla_dataset_path_info.contains_data_until,
            file_path=str(self.dataset_path),
            metadata_created_by=user_info.get_user_info_for_current_platform().short_email,
            subject_field=self._extract_subject_field_from_path(
                dapla_dataset_path_info,
            ),
            spatial_coverage_description=DEFAULT_SPATIAL_COVERAGE_DESCRIPTION,
        )
        metadata.variables = DatasetParser.for_file(dataset).get_fields()
        return metadata

    @staticmethod
    def build_metadata_document_path(
        dataset_path: ReadablePathLike,
    ) -> UPath:
        """Build the path to the metadata document corresponding to the given dataset.

        Args:
            dataset_path: Path to the dataset we wish to create metadata for.
        """
        dataset_path = UPath(dataset_path)
        return dataset_path.parent / (dataset_path.stem + METADATA_DOCUMENT_FILE_SUFFIX)

    def datadoc_model(self) -> all_optional_model.MetadataContainer:
        """Return the underlying datadoc model."""
        datadoc: ValidateDatadocMetadata = ValidateDatadocMetadata(
            percentage_complete=self.percent_complete,
            dataset=self.dataset,
            variables=self.variables,
        )
        if self.container:
            res = copy.deepcopy(self.container)
            res.datadoc = datadoc
            return res
        return all_optional_model.MetadataContainer(datadoc=datadoc)

    def write_metadata_document(self) -> None:
        """Write all currently known metadata to file.

        Side Effects:
            - Updates the dataset's metadata_last_updated_date and
                metadata_last_updated_by attributes.
            - Updates the dataset's file_path attribute.
            - Validates the metadata model and stores it in a MetadataContainer.
            - Writes the validated metadata to a file if the metadata_document
                attribute is set.
            - Logs the action and the content of the metadata document.

        Raises:
            ValueError: If no metadata document is specified for saving.
        """
        timestamp: datetime = get_timestamp_now()
        self.dataset.metadata_last_updated_date = timestamp
        self.dataset.metadata_last_updated_by = (
            user_info.get_user_info_for_current_platform().short_email
        )
        self.dataset.file_path = str(self.dataset_path)
        datadoc: ValidateDatadocMetadata = ValidateDatadocMetadata(
            percentage_complete=self.percent_complete,
            dataset=self.dataset,
            variables=self.variables,
        )
        if self.container:
            self.container.datadoc = datadoc
        else:
            self.container = all_optional_model.MetadataContainer(datadoc=datadoc)
        if self.metadata_document:
            content = self.container.model_dump_json(indent=4)
            self.metadata_document.write_text(content)
            logger.info("Saved metadata document %s", self.metadata_document)
            logger.info(
                "Metadata content",
                extra={"metadata_content": json.loads(content)},
            )
        else:
            msg = "No metadata document to save"
            raise ValueError(msg)

    @property
    def percent_complete(self) -> int:
        """The percentage of obligatory metadata completed.

        A metadata field is counted as complete when any non-None value is
        assigned. Used for a live progress bar in the UI, as well as being
        saved in the datadoc as a simple quality indicator.
        """
        num_all_fields = NUM_OBLIGATORY_DATASET_FIELDS + (
            NUM_OBLIGATORY_VARIABLES_FIELDS * len(self.variables)
        )
        num_set_fields = num_obligatory_dataset_fields_completed(
            self.dataset,
        ) + num_obligatory_variables_fields_completed(self.variables)
        return calculate_percentage(num_set_fields, num_all_fields)

    def add_pseudonymization(
        self,
        variable_short_name: str,
        pseudonymization: PseudonymizationType | None = None,
    ) -> None:
        """Adds a new pseudo variable to the list of pseudonymized variables.

        If `pseudonymization` is not supplied, an empty Pseudonymization structure
        will be created and assigned to the variable.
        If an encryption algorithm is recognized (one of the standard Dapla algorithms), default values are filled
        for any missing fields.

        Args:
            variable_short_name: The short name for the variable that one wants to update the pseudo for.
            pseudonymization: The updated pseudonymization.

        """
        variable = self.variables_lookup[variable_short_name]
        if pseudonymization:
            set_default_values_pseudonymization(variable, pseudonymization)
        else:
            if self.validate_required_fields_on_existing_metadata:
                msg = "Can't add empty pseudonymization object when validating required fields! Try setting `validate_required_fields_on_existing_metadata` to `False`."
                raise ValueError(msg)
            variable.pseudonymization = all_optional_model.Pseudonymization()

    def remove_pseudonymization(self, variable_short_name: str) -> None:
        """Removes a pseudo variable by using the shortname.

        Updates the pseudo variable lookup by creating a new one.

        Args:
            variable_short_name: The short name for the variable that one wants to remove the pseudo for.
        """
        if self.variables_lookup[variable_short_name].pseudonymization is not None:
            self.variables_lookup[variable_short_name].pseudonymization = None



"""Upgrade old metadata files to be compatible with new versions.

An important principle of Datadoc is that we ALWAYS guarantee backwards
compatibility of existing metadata documents. This means that we guarantee
that a user will never lose data, even if their document is decades old.

For each document version we release with breaking changes, we implement a
handler and register the version by defining a BackwardsCompatibleVersion
instance. These documents will then be upgraded when they're opened in Datadoc.

A test must also be implemented for each new version.
"""

from __future__ import annotations

import logging
from collections import OrderedDict
from dataclasses import dataclass
from typing import TYPE_CHECKING
from typing import Any

from dapla_metadata.datasets.compatibility._handlers import handle_current_version
from dapla_metadata.datasets.compatibility._handlers import handle_version_0_1_1
from dapla_metadata.datasets.compatibility._handlers import handle_version_1_0_0
from dapla_metadata.datasets.compatibility._handlers import handle_version_2_1_0
from dapla_metadata.datasets.compatibility._handlers import handle_version_2_2_0
from dapla_metadata.datasets.compatibility._handlers import handle_version_3_1_0
from dapla_metadata.datasets.compatibility._handlers import handle_version_3_2_0
from dapla_metadata.datasets.compatibility._handlers import handle_version_3_3_0
from dapla_metadata.datasets.compatibility._handlers import handle_version_4_0_0
from dapla_metadata.datasets.compatibility._handlers import handle_version_5_0_1
from dapla_metadata.datasets.compatibility._handlers import handle_version_6_0_0
from dapla_metadata.datasets.compatibility._utils import DATADOC_KEY
from dapla_metadata.datasets.compatibility._utils import DOCUMENT_VERSION_KEY
from dapla_metadata.datasets.compatibility._utils import UnknownModelVersionError
from dapla_metadata.datasets.compatibility._utils import (
    is_metadata_in_container_structure,
)

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from collections.abc import Callable

SUPPORTED_VERSIONS: OrderedDict[str, BackwardsCompatibleVersion] = OrderedDict()


@dataclass()
class BackwardsCompatibleVersion:
    """A version which we support with backwards compatibility.

    This class registers a version and its corresponding handler function
    for backwards compatibility.
    """

    version: str
    handler: Callable[[dict[str, Any]], dict[str, Any]]

    def __post_init__(self) -> None:
        """Register this version in the supported versions map.

        This method adds the instance to the `SUPPORTED_VERSIONS` dictionary
        using the version as the key.
        """
        SUPPORTED_VERSIONS[self.version] = self

    def upgrade(self, metadata: dict[str, Any]) -> dict[str, Any]:
        """Upgrade metadata from the format of the previous version to the format of this version.

        This method handles bumping the Document Version field so it's not necessary to do this in
        the individual handler functions.

        Args:
            metadata (dict[str, Any]): Metadata in the format of the previous version, to be upgraded.

        Returns:
            dict[str, Any]: The metadata upgraded to the version specified
        """
        metadata = self.handler(metadata)
        if is_metadata_in_container_structure(metadata):
            metadata[DATADOC_KEY][DOCUMENT_VERSION_KEY] = self.version
        else:
            metadata[DOCUMENT_VERSION_KEY] = self.version
        return metadata


# Register all the supported versions and their handlers.
BackwardsCompatibleVersion(version="0.1.1", handler=handle_version_0_1_1)
BackwardsCompatibleVersion(version="1.0.0", handler=handle_version_1_0_0)
BackwardsCompatibleVersion(version="2.1.0", handler=handle_version_2_1_0)
BackwardsCompatibleVersion(version="2.2.0", handler=handle_version_2_2_0)
BackwardsCompatibleVersion(version="3.1.0", handler=handle_version_3_1_0)
BackwardsCompatibleVersion(version="3.2.0", handler=handle_version_3_2_0)
BackwardsCompatibleVersion(version="3.3.0", handler=handle_version_3_3_0)
BackwardsCompatibleVersion(version="4.0.0", handler=handle_version_4_0_0)
BackwardsCompatibleVersion(version="5.0.1", handler=handle_version_5_0_1)
BackwardsCompatibleVersion(version="6.0.0", handler=handle_version_6_0_0)
BackwardsCompatibleVersion(version="6.1.0", handler=handle_current_version)


def upgrade_metadata(fresh_metadata: dict[str, Any]) -> dict[str, Any]:
    """Upgrade the metadata to the latest version using registered handlers.

    This function checks the version of the provided metadata and applies a series
    of upgrade handlers to migrate the metadata to the latest version.
    It starts from the provided version and applies all subsequent handlers in
    sequence. If the metadata is already in the latest version or the version
    cannot be determined, appropriate actions are taken.

    Args:
        fresh_metadata: The metadata dictionary to be upgraded. This dictionary
            must include version information that determines which handlers to apply.

    Returns:
        The upgraded metadata dictionary, after applying all necessary handlers.

    Raises:
        UnknownModelVersionError: If the metadata's version is unknown or unsupported.
    """
    if is_metadata_in_container_structure(fresh_metadata):
        if fresh_metadata[DATADOC_KEY] is None:
            return fresh_metadata
        supplied_version = fresh_metadata[DATADOC_KEY][DOCUMENT_VERSION_KEY]
    else:
        supplied_version = fresh_metadata[DOCUMENT_VERSION_KEY]
    start_running_handlers = False
    # Run all the handlers in order from the supplied version onwards
    for k, v in SUPPORTED_VERSIONS.items():
        if k == supplied_version:
            start_running_handlers = True
        if start_running_handlers:
            fresh_metadata = v.upgrade(fresh_metadata)
    if not start_running_handlers:
        raise UnknownModelVersionError(supplied_version)
    return fresh_metadata



import pytest
from klass import KlassClassification

from dapla_metadata.datasets.code_list import CodeList
from dapla_metadata.datasets.utility.enums import SupportedLanguages
from tests.datasets.constants import TEST_RESOURCES_DIRECTORY

CODE_LIST_DIR = "code_list"


@pytest.mark.parametrize(
    (
        "code_list_csv_filepath_nb",
        "code_list_csv_filepath_nn",
        "code_list_csv_filepath_en",
        "titles",
        "codes",
    ),
    [
        (
            TEST_RESOURCES_DIRECTORY / CODE_LIST_DIR / "code_list_nb.csv",
            TEST_RESOURCES_DIRECTORY / CODE_LIST_DIR / "code_list_nn.csv",
            TEST_RESOURCES_DIRECTORY / CODE_LIST_DIR / "code_list_en.csv",
            ["Adresse", "Arbeidsulykke", "Bolig"],
            ["01", "02", "03"],
        ),
        (
            TEST_RESOURCES_DIRECTORY / CODE_LIST_DIR / "no_code.csv",
            TEST_RESOURCES_DIRECTORY / CODE_LIST_DIR / "no_code.csv",
            TEST_RESOURCES_DIRECTORY / CODE_LIST_DIR / "no_code.csv",
            ["Adresse", "Arbeidsulykke", "Bolig"],
            [None, None, None],
        ),
    ],
)
@pytest.mark.usefixtures("_mock_fetch_dataframe")
def test_read_dataframe(
    code_list_fake_structure: CodeList,
    codes: list[str],
    titles: list[str],
):
    code_list_fake_structure.wait_for_external_result()

    for i in range(3):
        assert code_list_fake_structure.classifications[i].code == codes[i]

    for idx in range(3):
        classification = code_list_fake_structure.classifications[idx]
        assert classification.get_title(SupportedLanguages.NORSK_BOKMÅL) == titles[idx]
        assert classification.get_title(SupportedLanguages.ENGLISH) == titles[idx]
        assert classification.get_title(SupportedLanguages.NORSK_NYNORSK) == titles[idx]


def test_non_existent_code(thread_pool_executor):
    code_list = CodeList(thread_pool_executor, 0)
    code_list.wait_for_external_result()
    assert code_list.classifications == []


@pytest.mark.parametrize("language", list(SupportedLanguages))
def test_instantiate_classification(language: SupportedLanguages):
    # This started failing when casting the enum language argument
    # to a str, so we write a test to cover this eventuality.
    assert isinstance(
        KlassClassification("6", language.lower()),  # type: ignore [arg-type]
        KlassClassification,
    )



"""Code relating to merging metadata from an existing metadata document and metadata extracted from a new dataset.

This is primarily convenience functionality for users whereby they can programmatically generate metadata without
having to manually enter it. This is primarily useful when data is sharded by time (i.e. each dataset applies for
a particular period like a month or a year). Assuming there aren't structural changes, the metadata may be reused
for all periods.

It is important to be able to detect changes in the structure of the data and warn users about this so that they can
make changes as appropriate.
"""

import copy
import logging
import warnings
from collections.abc import Iterable
from dataclasses import dataclass
from dataclasses import field
from pathlib import Path
from typing import cast

import datadoc_model
import datadoc_model.all_optional.model as all_optional_model
import datadoc_model.required.model as required_model
from upath import UPath

from dapla_metadata.datasets.dapla_dataset_path_info import DaplaDatasetPathInfo
from dapla_metadata.datasets.utility.constants import (
    DATASET_FIELDS_FROM_EXISTING_METADATA,
)
from dapla_metadata.datasets.utility.constants import INCONSISTENCIES_MESSAGE
from dapla_metadata.datasets.utility.utils import OptionalDatadocMetadataType
from dapla_metadata.datasets.utility.utils import VariableListType

logger = logging.getLogger(__name__)

BUCKET_NAME_MESSAGE = "Bucket name"
DATA_PRODUCT_NAME_MESSAGE = "Data product name"
DATASET_STATE_MESSAGE = "Dataset state"
DATASET_SHORT_NAME_MESSAGE = "Dataset short name"
VARIABLES_ADDITIONAL_MESSAGE = (
    "Dataset has additional variables than defined in metadata"
)
VARIABLE_RENAME_MESSAGE = "Variables have been renamed in the dataset"
VARIABLE_ORDER_MESSAGE = "The order of variables in the dataset has changed"
VARIABLE_DATATYPES_MESSAGE = "Variable datatypes differ"
VARIABLES_FEWER_MESSAGE = "Dataset has fewer variables than defined in metadata"


class InconsistentDatasetsWarning(UserWarning):
    """Existing and new datasets differ significantly from one another."""


class InconsistentDatasetsError(ValueError):
    """Existing and new datasets differ significantly from one another."""


@dataclass
class DatasetConsistencyStatus:
    """Store the status for different aspects of dataset consistency.

    Attributes:
        message: Communicates to the user what aspect is inconsistent.
        success: False if inconsistency is detected.
        variables: Optionally communicate which variables are affected.
    """

    message: str
    success: bool
    variables: Iterable[str] = field(default_factory=list)

    def __str__(self) -> str:
        """Format the user message."""
        message = self.message
        if self.variables:
            message += f"\n\tVariables: {self.variables}"
        return message


def check_dataset_consistency(
    new_dataset_path: Path | UPath,
    existing_dataset_path: Path | UPath,
) -> list[DatasetConsistencyStatus]:
    """Run consistency tests.

    Args:
        new_dataset_path: Path to the dataset to be documented.
        existing_dataset_path: Path stored in the existing metadata.

    Returns:
        List of consistency check results.
    """
    new_dataset_path_info = DaplaDatasetPathInfo(new_dataset_path)
    existing_dataset_path_info = DaplaDatasetPathInfo(existing_dataset_path)
    return [
        DatasetConsistencyStatus(
            message=BUCKET_NAME_MESSAGE,
            success=(
                new_dataset_path_info.bucket_name
                == existing_dataset_path_info.bucket_name
            ),
        ),
        DatasetConsistencyStatus(
            message=DATA_PRODUCT_NAME_MESSAGE,
            success=(
                new_dataset_path_info.statistic_short_name
                == existing_dataset_path_info.statistic_short_name
            ),
        ),
        DatasetConsistencyStatus(
            message=DATASET_STATE_MESSAGE,
            success=(
                new_dataset_path_info.dataset_state
                == existing_dataset_path_info.dataset_state
            ),
        ),
        DatasetConsistencyStatus(
            message=DATASET_SHORT_NAME_MESSAGE,
            success=(
                new_dataset_path_info.dataset_short_name
                == existing_dataset_path_info.dataset_short_name
            ),
        ),
    ]


def check_variables_consistency(
    extracted_variables: VariableListType,
    existing_variables: VariableListType,
) -> list[DatasetConsistencyStatus]:
    """Check for consistency in variables structure.

    Compares the existing metadata and that extracted from the new dataset and provides
    highly detailed feedback on what is different between them.

    We don't return all the results because that could create conflicting messages and false positives.

    Args:
        extracted_variables (VariableListType): Variables extracted from the new dataset.
        existing_variables (VariableListType): Variables already documented in existing metadata

    Returns:
        list[DatasetConsistencyStatus]: The list of checks and whether they were successful.
    """
    extracted_names_set = {v.short_name or "" for v in extracted_variables}
    existing_names_set = {v.short_name or "" for v in existing_variables}
    same_length = len(extracted_variables) == len(existing_variables)
    more_extracted_variables = extracted_names_set.difference(existing_names_set)
    fewer_extracted_variables = existing_names_set.difference(extracted_names_set)
    results = []
    if same_length:
        if more_extracted_variables:
            results.append(
                DatasetConsistencyStatus(
                    message=VARIABLE_RENAME_MESSAGE,
                    variables=more_extracted_variables,
                    success=not bool(more_extracted_variables),
                )
            )
        else:
            results.append(
                DatasetConsistencyStatus(
                    message=VARIABLE_ORDER_MESSAGE,
                    success=[v.short_name or "" for v in extracted_variables]
                    == [v.short_name or "" for v in existing_variables],
                )
            )
            results.append(
                DatasetConsistencyStatus(
                    message=VARIABLE_DATATYPES_MESSAGE,
                    success=[v.data_type for v in extracted_variables]
                    == [v.data_type for v in existing_variables],
                )
            )
    else:
        results.extend(
            [
                DatasetConsistencyStatus(
                    message=VARIABLES_ADDITIONAL_MESSAGE,
                    variables=more_extracted_variables,
                    success=not bool(more_extracted_variables),
                ),
                DatasetConsistencyStatus(
                    message=VARIABLES_FEWER_MESSAGE,
                    variables=fewer_extracted_variables,
                    success=not bool(fewer_extracted_variables),
                ),
            ]
        )
    return results


def check_ready_to_merge(
    results: list[DatasetConsistencyStatus], *, errors_as_warnings: bool
) -> None:
    """Check if the datasets are consistent enough to make a successful merge of metadata.

    Args:
        results: List if dict with property name and boolean success flag
        errors_as_warnings: True if failing checks should be raised as warnings, not errors.

    Raises:
        InconsistentDatasetsError: If inconsistencies are found and `errors_as_warnings == False`
    """
    if failures := [result for result in results if not result.success]:
        messages_list = "\n - ".join(str(f) for f in failures)
        msg = f"{INCONSISTENCIES_MESSAGE}\n - {messages_list}"
        if errors_as_warnings:
            warnings.warn(
                message=msg,
                category=InconsistentDatasetsWarning,
                stacklevel=2,
            )
        else:
            raise InconsistentDatasetsError(
                msg,
            )


def override_dataset_fields(
    merged_metadata: all_optional_model.DatadocMetadata,
    existing_metadata: all_optional_model.DatadocMetadata
    | required_model.DatadocMetadata,
) -> None:
    """Overrides specific fields in the dataset of `merged_metadata` with values from the dataset of `existing_metadata`.

    This function iterates over a predefined list of fields, `DATASET_FIELDS_FROM_EXISTING_METADATA`,
    and sets the corresponding fields in the `merged_metadata.dataset` object to the values
    from the `existing_metadata.dataset` object.

    Args:
        merged_metadata: An instance of `DatadocMetadata` containing the dataset to be updated.
        existing_metadata: An instance of `DatadocMetadata` containing the dataset whose values are used to update `merged_metadata.dataset`.

    Returns:
        `None`.
    """
    if merged_metadata.dataset and existing_metadata.dataset:
        # Override the fields as defined
        for field in DATASET_FIELDS_FROM_EXISTING_METADATA:
            setattr(
                merged_metadata.dataset,
                field,
                getattr(existing_metadata.dataset, field),
            )


def merge_variables(
    existing_metadata: OptionalDatadocMetadataType,
    extracted_metadata: all_optional_model.DatadocMetadata,
    merged_metadata: all_optional_model.DatadocMetadata,
) -> all_optional_model.DatadocMetadata:
    """Merges variables from the extracted metadata into the existing metadata and updates the merged metadata.

    This function compares the variables from `extracted_metadata` with those in `existing_metadata`.
    For each variable in `extracted_metadata`, it checks if a variable with the same `short_name` exists
    in `existing_metadata`. If a match is found, it updates the existing variable with information from
    `extracted_metadata`. If no match is found, the variable from `extracted_metadata` is directly added to `merged_metadata`.

    Args:
        existing_metadata: The metadata object containing the current state of variables.
        extracted_metadata: The metadata object containing new or updated variables to merge.
        merged_metadata: The metadata object that will contain the result of the merge.

    Returns:
        all_optional_model.DatadocMetadata: The `merged_metadata` object containing variables from both `existing_metadata`
        and `extracted_metadata`.
    """
    if (
        existing_metadata is not None
        and existing_metadata.variables is not None
        and extracted_metadata is not None
        and extracted_metadata.variables is not None
        and merged_metadata.variables is not None
    ):
        for extracted in extracted_metadata.variables:
            existing = next(
                (
                    existing
                    for existing in existing_metadata.variables
                    if existing.short_name == extracted.short_name
                ),
                None,
            )
            if existing:
                existing.id = (
                    None  # Set to None so that it will be set assigned a fresh ID later
                )
                existing.contains_data_from = (
                    extracted.contains_data_from or existing.contains_data_from
                )
                existing.contains_data_until = (
                    extracted.contains_data_until or existing.contains_data_until
                )
                merged_metadata.variables.append(
                    cast("datadoc_model.all_optional.model.Variable", existing)
                )
            else:
                # If there is no existing metadata for this variable, we just use what we have extracted
                merged_metadata.variables.append(extracted)
    return merged_metadata


def merge_metadata(
    extracted_metadata: all_optional_model.DatadocMetadata | None,
    existing_metadata: OptionalDatadocMetadataType,
) -> all_optional_model.DatadocMetadata:
    if not existing_metadata:
        logger.warning(
            "No existing metadata found, no merge to perform. Continuing with extracted metadata.",
        )
        return extracted_metadata or all_optional_model.DatadocMetadata()

    if not extracted_metadata:
        return cast("all_optional_model.DatadocMetadata", existing_metadata)

    # Use the extracted metadata as a base
    merged_metadata = all_optional_model.DatadocMetadata(
        dataset=copy.deepcopy(extracted_metadata.dataset),
        variables=[],
    )

    override_dataset_fields(
        merged_metadata=merged_metadata,
        existing_metadata=existing_metadata,
    )

    # Merge variables.
    # For each extracted variable, copy existing metadata into the merged metadata
    return merge_variables(
        existing_metadata=existing_metadata,
        extracted_metadata=extracted_metadata,
        merged_metadata=merged_metadata,
    )



import urllib.parse
from pathlib import Path

import ruamel.yaml
from testcontainers.core.image import DockerImage
from testcontainers.core.wait_strategies import LogMessageWaitStrategy
from testcontainers.generic import ServerContainer


class MicrocksContainer(ServerContainer):
    """Microcks Container.

    [Microcks](https://microcks.io/) allows mocking arbitrary API services based on an OpenAPI definition.

    This implementation supports a subset of Microcks functionality, supporting mocking a single service
    by uploading a single local definition file.
    """

    def __init__(
        self,
        port: int = 8080,
        image: str | DockerImage = "quay.io/microcks/microcks-uber:1.10.1-native",
    ) -> None:
        """Instantiate a Microcks container.

        Args:
            port (int, optional): The internal application port to use. Defaults to 8080.
            image (str | DockerImage, optional): The image name. Must be compatible with quay.io/microcks/microcks-uber. Defaults to "quay.io/microcks/microcks-uber:1.10.1-native".
        """
        super().__init__(port, image)
        self.primary_artifact: str | None = None

    def start(self) -> "ServerContainer":
        """Start the container and wait for it to be ready."""
        started = super().start()
        started.waiting_for(
            LogMessageWaitStrategy(
                message=".*Started MicrocksApplication.*"
            ).with_startup_timeout(10)
        )
        return started

    def get_api_url(self) -> str:
        """Get the base URL for the container."""
        return self._create_connection_url()

    def get_mock_url(self) -> str | None:
        """Get the base URL for the mocked service.

        Only relevant if an artifact has been uploaded.

        Returns:
            str | None: The base URL for the primary artifact, or None.
        """
        if not self.primary_artifact:
            return None

        version: str | None = None
        api_name: str | None = None

        with Path(self.primary_artifact).open() as openapi:
            openapi_yaml: dict = ruamel.yaml.YAML().load(openapi)

        info: dict | None = openapi_yaml.get("info")
        if info is None:
            return None
        version = info.get("version")
        api_name = info.get("title")

        if not version or not api_name:
            return None

        return (
            f"{self.get_api_url()}/rest/{urllib.parse.quote_plus(api_name)}/{version}"
        )

    def upload_primary_artifact(self, openapi_definition_path: str) -> str:
        """Upload a local OpenAPI definition file as the primary artifact.

        Args:
            openapi_definition_path (str): Path to the OpenAPI definition file

        Returns:
            _type_: _description_
        """
        with Path(openapi_definition_path).open("rb") as openapi:
            upload_response = self.get_client().post(
                "/api/artifact/upload",
                files={"file": openapi},
                timeout=30,
            )
            upload_response.raise_for_status()
        self.primary_artifact = openapi_definition_path
        return self.primary_artifact



from datetime import date

import pytest
from pydantic import BaseModel
from pydantic import ValidationError

from dapla_metadata.variable_definitions._generated.vardef_client.models.complete_response import (
    CompleteResponse,
)
from dapla_metadata.variable_definitions._generated.vardef_client.models.contact import (
    Contact,
)
from dapla_metadata.variable_definitions._generated.vardef_client.models.draft import (
    Draft,
)
from dapla_metadata.variable_definitions._generated.vardef_client.models.language_string_type import (
    LanguageStringType,
)
from dapla_metadata.variable_definitions._generated.vardef_client.models.owner import (
    Owner,
)
from dapla_metadata.variable_definitions._generated.vardef_client.models.patch import (
    Patch,
)
from dapla_metadata.variable_definitions._generated.vardef_client.models.update_draft import (
    UpdateDraft,
)
from dapla_metadata.variable_definitions._generated.vardef_client.models.validity_period import (
    ValidityPeriod,
)
from dapla_metadata.variable_definitions._generated.vardef_client.models.variable_status import (
    VariableStatus,
)

# NOTE: If we need to change these tests, that will normally mean it's a BREAKING CHANGE for users


@pytest.mark.parametrize(
    "model",
    [
        Draft,
        CompleteResponse,
        Contact,
        Owner,
        ValidityPeriod,
    ],
)
def test_empty_instantiation_model_with_required_fields(model: type[BaseModel]):
    with pytest.raises(ValidationError):
        model()


@pytest.mark.parametrize(
    "model",
    [
        LanguageStringType,
        Patch,
        UpdateDraft,
    ],
)
def test_empty_instantiation_model_without_required_fields(model: type[BaseModel]):
    model()


def test_instantiate_draft_all_fields(language_string_type, contact):
    Draft(
        name=language_string_type,
        short_name="test",
        definition=language_string_type,
        classification_reference="91",
        unit_types=["a", "b"],
        subject_fields=["a", "b"],
        contains_special_categories_of_personal_data=True,
        measurement_type="test",
        valid_from=date(2024, 11, 1),
        external_reference_uri="http://www.example.com",
        comment=language_string_type,
        related_variable_definition_uris=["http://www.example.com"],
        contact=contact,
    )


def test_instantiate_update_draft_all_fields(language_string_type, contact, owner):
    UpdateDraft(
        name=language_string_type,
        short_name="test",
        definition=language_string_type,
        classification_reference="91",
        unit_types=["a", "b"],
        subject_fields=["a", "b"],
        contains_special_categories_of_personal_data=True,
        variable_status=VariableStatus.PUBLISHED_EXTERNAL,
        measurement_type="test",
        valid_from=date(2024, 11, 1),
        external_reference_uri="http://www.example.com",
        comment=language_string_type,
        related_variable_definition_uris=["http://www.example.com"],
        contact=contact,
        owner=owner,
    )


def test_instantiate_patch_all_fields(language_string_type, contact, owner):
    Patch(
        name=language_string_type,
        definition=language_string_type,
        classification_reference="91",
        unit_types=["a", "b"],
        subject_fields=["a", "b"],
        contains_special_categories_of_personal_data=True,
        variable_status=VariableStatus.PUBLISHED_EXTERNAL,
        measurement_type="test",
        valid_until=date(2024, 11, 1),
        external_reference_uri="http://www.example.com",
        comment=language_string_type,
        related_variable_definition_uris=["http://www.example.com"],
        contact=contact,
        owner=owner,
    )


def test_instantiate_validity_period_all_fields(language_string_type, contact):
    ValidityPeriod(
        name=language_string_type,
        definition=language_string_type,
        classification_reference="91",
        unit_types=["a", "b"],
        subject_fields=["a", "b"],
        contains_special_categories_of_personal_data=True,
        variable_status=VariableStatus.PUBLISHED_EXTERNAL,
        measurement_type="test",
        valid_from=date(2024, 11, 1),
        external_reference_uri="http://www.example.com",
        comment=language_string_type,
        related_variable_definition_uris=["http://www.example.com"],
        contact=contact,
    )



import datetime

import pytest

from dapla_metadata.datasets.utility.utils import calculate_percentage
from dapla_metadata.datasets.utility.utils import get_current_date
from dapla_metadata.datasets.utility.utils import incorrect_date_order
from dapla_metadata.datasets.utility.utils import running_in_notebook


def test_calculate_percentage():
    assert calculate_percentage(1, 3) == 33


@pytest.mark.parametrize(
    ("date_from", "date_until", "expected"),
    [
        (datetime.date(2024, 1, 1), datetime.date(1960, 1, 1), True),
        (datetime.date(1980, 1, 1), datetime.date(2000, 6, 5), False),
        (None, None, False),
        (datetime.date(2024, 1, 1), None, False),
        (None, datetime.date(2024, 1, 1), True),
        (datetime.date(2024, 1, 1), datetime.date(2024, 1, 1), False),
    ],
)
def test_incorrect_date_order(date_from, date_until, expected):
    result = incorrect_date_order(date_from, date_until)
    assert result == expected


def test_not_running_in_notebook():
    assert not running_in_notebook()


def test_get_current_date():
    date = get_current_date()
    assert isinstance(date, str)
    datetime.datetime.strptime(date, "%Y-%m-%d").date()  # noqa: DTZ007



from enum import Enum


class DaplaRegion(str, Enum):
    """Dapla platforms/regions."""

    DAPLA_LAB = "DAPLA_LAB"
    ON_PREM = "ON_PREM"
    CLOUD_RUN = "CLOUD_RUN"


class DaplaEnvironment(str, Enum):
    """Dapla lifecycle environment."""

    PROD = "PROD"
    TEST = "TEST"
    DEV = "DEV"


class DaplaService(str, Enum):
    """Dapla services."""

    DATADOC = "DATADOC"
    JUPYTERLAB = "JUPYTERLAB"
    VS_CODE = "VS_CODE"
    R_STUDIO = "R_STUDIO"
    KILDOMATEN = "KILDOMATEN"



"""Simple tests for basic coverage of the generated client."""

from dapla_metadata.variable_definitions._generated import vardef_client
from dapla_metadata.variable_definitions._generated.vardef_client.models.complete_response import (
    CompleteResponse,
)
from tests.utils.constants import VARDEF_EXAMPLE_DATE
from tests.utils.constants import VARDEF_EXAMPLE_DEFINITION_ID


def test_create_draft(api_client, draft):
    api_instance = vardef_client.DraftVariableDefinitionsApi(api_client)
    definition = api_instance.create_variable_definition(
        draft=draft,
    )
    assert isinstance(definition, CompleteResponse)


def test_update_draft(api_client, update_draft):
    api_instance = vardef_client.DraftVariableDefinitionsApi(api_client)
    definition = api_instance.update_variable_definition_by_id(
        variable_definition_id=VARDEF_EXAMPLE_DEFINITION_ID,
        update_draft=update_draft,
    )
    assert isinstance(definition, CompleteResponse)


def test_delete_draft(api_client):
    api_instance = vardef_client.DraftVariableDefinitionsApi(api_client)
    response = api_instance.delete_variable_definition_by_id(
        variable_definition_id=VARDEF_EXAMPLE_DEFINITION_ID,
    )
    assert response is None


def test_list_variable_definitions(api_client):
    api_instance = vardef_client.VariableDefinitionsApi(api_client)
    definitions = api_instance.list_variable_definitions(
        date_of_validity=VARDEF_EXAMPLE_DATE,
    )
    assert isinstance(definitions[0], CompleteResponse)


def test_get_variable_definition(api_client):
    api_instance = vardef_client.VariableDefinitionsApi(api_client)
    definition = api_instance.get_variable_definition_by_id(
        variable_definition_id=VARDEF_EXAMPLE_DEFINITION_ID,
        date_of_validity=VARDEF_EXAMPLE_DATE,
    )
    assert isinstance(definition, CompleteResponse)


def test_list_validity_periods(api_client):
    api_instance = vardef_client.ValidityPeriodsApi(api_client)
    validity_periods = api_instance.list_validity_periods(
        variable_definition_id=VARDEF_EXAMPLE_DEFINITION_ID,
    )
    assert isinstance(validity_periods[0], CompleteResponse)


def test_create_validity_period(api_client, validity_period):
    api_instance = vardef_client.ValidityPeriodsApi(api_client)
    validity_period = api_instance.create_validity_period(
        variable_definition_id=VARDEF_EXAMPLE_DEFINITION_ID,
        validity_period=validity_period,
    )
    assert isinstance(validity_period, CompleteResponse)


def test_list_patches(api_client):
    api_instance = vardef_client.PatchesApi(api_client)
    patches = api_instance.list_patches(
        variable_definition_id=VARDEF_EXAMPLE_DEFINITION_ID,
    )
    assert isinstance(patches[0], CompleteResponse)


def test_get_patch(api_client):
    api_instance = vardef_client.PatchesApi(api_client)
    patch = api_instance.get_patch(
        variable_definition_id=VARDEF_EXAMPLE_DEFINITION_ID,
        patch_id=1,
    )
    assert isinstance(patch, CompleteResponse)


def test_create_patch(api_client, patch_fixture):
    api_instance = vardef_client.PatchesApi(api_client)
    patch = api_instance.create_patch(
        variable_definition_id=VARDEF_EXAMPLE_DEFINITION_ID,
        valid_from=VARDEF_EXAMPLE_DATE,
        patch=patch_fixture,
    )
    assert isinstance(patch, CompleteResponse)



import contextlib
from pathlib import Path

import fsspec  # type: ignore  # noqa: PGH003
import pytest
from datadoc_model.all_optional.model import DatadocMetadata
from datadoc_model.all_optional.model import DataType
from datadoc_model.all_optional.model import Variable
from fsspec.registry import register_implementation  # type: ignore[import-untyped]
from upath import UPath

from dapla_metadata.datasets._merge import BUCKET_NAME_MESSAGE
from dapla_metadata.datasets._merge import DATA_PRODUCT_NAME_MESSAGE
from dapla_metadata.datasets._merge import DATASET_SHORT_NAME_MESSAGE
from dapla_metadata.datasets._merge import DATASET_STATE_MESSAGE
from dapla_metadata.datasets._merge import VARIABLE_DATATYPES_MESSAGE
from dapla_metadata.datasets._merge import VARIABLE_ORDER_MESSAGE
from dapla_metadata.datasets._merge import VARIABLE_RENAME_MESSAGE
from dapla_metadata.datasets._merge import VARIABLES_ADDITIONAL_MESSAGE
from dapla_metadata.datasets._merge import VARIABLES_FEWER_MESSAGE
from dapla_metadata.datasets._merge import DatasetConsistencyStatus
from dapla_metadata.datasets._merge import InconsistentDatasetsError
from dapla_metadata.datasets._merge import InconsistentDatasetsWarning
from dapla_metadata.datasets._merge import check_dataset_consistency
from dapla_metadata.datasets._merge import check_ready_to_merge
from dapla_metadata.datasets._merge import check_variables_consistency
from dapla_metadata.datasets.core import Datadoc
from tests.datasets.constants import TEST_BUCKET_NAMING_STANDARD_COMPATIBLE_PATH
from tests.datasets.constants import VARIABLE_DATA_TYPES
from tests.datasets.constants import VARIABLE_SHORT_NAMES


@pytest.mark.parametrize(
    ("new_dataset_path", "existing_dataset_path"),
    [
        (
            TEST_BUCKET_NAMING_STANDARD_COMPATIBLE_PATH,
            TEST_BUCKET_NAMING_STANDARD_COMPATIBLE_PATH,
        ),
        (
            TEST_BUCKET_NAMING_STANDARD_COMPATIBLE_PATH.replace("v1", "v2"),
            TEST_BUCKET_NAMING_STANDARD_COMPATIBLE_PATH,
        ),
        (
            TEST_BUCKET_NAMING_STANDARD_COMPATIBLE_PATH.replace("p2021", "p2022"),
            TEST_BUCKET_NAMING_STANDARD_COMPATIBLE_PATH,
        ),
        (
            TEST_BUCKET_NAMING_STANDARD_COMPATIBLE_PATH.replace(
                "/ifpn",
                "/deeper/folder/structure/ifpn",
            ),
            TEST_BUCKET_NAMING_STANDARD_COMPATIBLE_PATH,
        ),
    ],
    ids=[
        "identical path",
        "differing version",
        "differing period",
        "different folder structure",
    ],
)
def test_check_dataset_consistency_consistent_paths(
    new_dataset_path: str,
    existing_dataset_path: str,
):
    result = check_dataset_consistency(
        UPath(new_dataset_path),
        UPath(existing_dataset_path),
    )
    for r in result:
        assert r.success, f"'{r.message}' failed"


@pytest.mark.parametrize(
    ("new_dataset_path", "existing_dataset_path"),
    [
        (
            TEST_BUCKET_NAMING_STANDARD_COMPATIBLE_PATH.replace("produkt", "delt"),
            TEST_BUCKET_NAMING_STANDARD_COMPATIBLE_PATH,
        ),
        (
            TEST_BUCKET_NAMING_STANDARD_COMPATIBLE_PATH.replace("ifpn", "blah"),
            TEST_BUCKET_NAMING_STANDARD_COMPATIBLE_PATH,
        ),
        (
            TEST_BUCKET_NAMING_STANDARD_COMPATIBLE_PATH.replace(
                "klargjorte_data",
                "utdata",
            ),
            TEST_BUCKET_NAMING_STANDARD_COMPATIBLE_PATH,
        ),
        (
            TEST_BUCKET_NAMING_STANDARD_COMPATIBLE_PATH.replace(
                "person_testdata",
                "totally_different_dataset",
            ),
            TEST_BUCKET_NAMING_STANDARD_COMPATIBLE_PATH,
        ),
    ],
    ids=["bucket name", "data product name", "dataset state", "dataset short name"],
)
def test_check_dataset_consistency_inconsistent_paths(
    new_dataset_path: str, existing_dataset_path: str, request
):
    result = check_dataset_consistency(
        UPath(new_dataset_path),
        UPath(existing_dataset_path),
    )
    test_id = request.node.callspec.id
    result_entry = next(r for r in result if r.message.lower() == test_id)  # type: ignore[attr-defined]
    assert not result_entry.success


@pytest.mark.parametrize(
    ("dataset_consistency_status"),
    [
        [
            DatasetConsistencyStatus(
                message=BUCKET_NAME_MESSAGE,
                success=False,
            ),
            DatasetConsistencyStatus(
                message=DATA_PRODUCT_NAME_MESSAGE,
                success=True,
            ),
            DatasetConsistencyStatus(
                message=DATASET_STATE_MESSAGE,
                success=True,
            ),
            DatasetConsistencyStatus(
                message=DATASET_SHORT_NAME_MESSAGE,
                success=True,
            ),
        ],
    ],
)
@pytest.mark.parametrize(
    "errors_as_warnings",
    [True, False],
    ids=["warnings", "errors"],
)
def test_check_ready_to_merge_errors_as_warnings(
    dataset_consistency_status: list[DatasetConsistencyStatus],
    errors_as_warnings: bool,
):
    with contextlib.ExitStack() as stack:
        if errors_as_warnings:
            stack.enter_context(pytest.warns(InconsistentDatasetsWarning))
        else:
            stack.enter_context(pytest.raises(InconsistentDatasetsError))
        check_ready_to_merge(
            dataset_consistency_status,
            errors_as_warnings=errors_as_warnings,
        )


def test_check_dataset_consistency_inconsistent_variable_data_types():
    def create_variables(data_types):
        return [
            Variable(short_name=name, data_type=data_type)
            for name, data_type in zip(VARIABLE_SHORT_NAMES, data_types, strict=False)
        ]

    metadata1 = DatadocMetadata(
        variables=create_variables([*VARIABLE_DATA_TYPES[:-1], DataType.BOOLEAN])
    )
    metadata2 = DatadocMetadata(variables=create_variables(VARIABLE_DATA_TYPES))
    result = check_variables_consistency(
        metadata1.variables,
        metadata2.variables,
    )
    for r in result:
        if VARIABLE_DATATYPES_MESSAGE in r.message:
            assert not r.success, f"'{r.message}' passed but should have failed"
        else:
            assert r.success, f"'{r.message}' failed but should have passed"


@pytest.mark.parametrize(
    ("extracted_variable_names", "existing_variable_names", "expected_message"),
    [
        (
            VARIABLE_SHORT_NAMES,
            VARIABLE_SHORT_NAMES[:-2],
            VARIABLES_ADDITIONAL_MESSAGE,
        ),
        (
            VARIABLE_SHORT_NAMES[:-2],
            VARIABLE_SHORT_NAMES,
            VARIABLES_FEWER_MESSAGE,
        ),
        (
            VARIABLE_SHORT_NAMES,
            [*VARIABLE_SHORT_NAMES[:-1], "blah"],
            VARIABLE_RENAME_MESSAGE,
        ),
        (
            sorted(VARIABLE_SHORT_NAMES),
            VARIABLE_SHORT_NAMES,
            VARIABLE_ORDER_MESSAGE,
        ),
        (
            VARIABLE_SHORT_NAMES,
            VARIABLE_SHORT_NAMES,
            None,
        ),
    ],
    ids=[
        "additional extracted",
        "fewer extracted",
        "renamed",
        "order changed",
        "identical",
    ],
)
def test_check_variables_consistency(
    extracted_variable_names: list[str],
    existing_variable_names: list[str],
    expected_message: str | None,
):
    results = check_variables_consistency(
        [Variable(short_name=name) for name in extracted_variable_names],
        [Variable(short_name=name) for name in existing_variable_names],
    )
    if expected_message:
        for r in results:
            if expected_message in r.message:
                assert not r.success, f"'{r.message}' passed but should have failed"
            else:
                assert r.success, f"'{r.message}' failed but should have passed"
    else:
        assert all(r.success for r in results)


def test_bucket_check_ok_when_both_params_are_gs_paths(mocker):
    """Test the functionality of `check_dataset_consistency`.

    Verify the desired behavior when BOTH parameters are gs-paths (no Path() conversion occurs). In
    this case no `Bucket name` warning should be issued.
    """
    _patch_gs_filesystem(mocker)
    resource_json = (
        Path(__file__).parent
        / "resources"
        / "existing_metadata_file"
        / "person_testdata_p2020-12-31_p2020-12-31_v1__DOC.json"
    )
    resource_parquet = (
        Path(__file__).parent
        / "resources"
        / "datasets"
        / "person_testdata_p2021-12-31_p2021-12-31_v1.parquet"
    )
    assert resource_parquet.exists(), f"Parquet resource missing: {resource_parquet}"
    assert resource_json.exists(), f"JSON resource missing: {resource_json}"
    directory = "ssb-dapla-felles-data-produkt-test/datadoc/brukertest/10/sykefratot/klargjorte_data/"
    dataset_gs = UPath(
        f"gs://{directory}person_testdata_p2021-12-31_p2021-12-31_v1.parquet"
    )
    metadata_doc_gs = UPath(
        f"gs://{directory}person_testdata_p2021-12-31_p2021-12-31_v1__DOC.json"
    )
    metadata_doc_gs.parent.mkdir(parents=True, exist_ok=True)
    metadata_doc_gs.write_text(resource_json.read_text(), encoding="utf-8")
    assert metadata_doc_gs.exists(), (
        "Metadata document was not written to memory filesystem"
    )
    dataset_gs.parent.mkdir(parents=True, exist_ok=True)
    dataset_gs.write_bytes(resource_parquet.read_bytes())
    assert dataset_gs.exists(), "Parquet file was not written to memory filesystem"
    dd = Datadoc(
        dataset_path=str(dataset_gs),
        metadata_document_path=str(metadata_doc_gs),
        errors_as_warnings=True,
    )
    _assert_bucket_ok(dd.dataset_consistency_status, BUCKET_NAME_MESSAGE)


def _assert_bucket_ok(results: list, expected_msg: str):
    bucket_checks = [r for r in results if expected_msg in r.message]
    assert bucket_checks, f"Expected at least one check for {expected_msg}."
    assert all(r.success for r in bucket_checks), (
        f"Expected {expected_msg} to have success == True when both are gs-paths."
    )


def _patch_gs_filesystem(mocker):
    """Mock GCS filesystem to use memory filesystem for testing.

    This allows us to use gs:// paths with UPath while using an in-memory
    filesystem that doesn't require actual GCS authentication or connectivity.
    """
    memory_fs = fsspec.filesystem("memory")

    original_filesystem = fsspec.filesystem

    def mock_filesystem(protocol, **kwargs):
        """Return memory filesystem for gs:// protocol, otherwise use original."""
        if protocol == "gs":
            return memory_fs
        return original_filesystem(protocol, **kwargs)

    mocker.patch("fsspec.filesystem", side_effect=mock_filesystem)

    register_implementation("gs", memory_fs.__class__, "MemoryFileSystem")

    def mock_fs_factory(_urlpath, protocol, storage_options):
        """Return memory filesystem for gs:// protocol."""
        if protocol == "gs":
            return memory_fs
        return fsspec.filesystem(protocol, **storage_options)

    mocker.patch("upath.core.UPath._fs_factory", side_effect=mock_fs_factory)



"""Script to upgrade metadata documents to the newest version."""

import argparse

from upath import UPath

from dapla_metadata.datasets.core import Datadoc

parser = argparse.ArgumentParser(
    description="Upgrade metadata documents to the newest version"
)
parser.add_argument(
    "--path",
    nargs="+",
    type=lambda p: UPath(p).absolute(),
    help="Path to a metadata document to upgrade",
)
args = parser.parse_args()

for doc in args.path:
    meta = Datadoc(metadata_document_path=str(doc))
    meta.write_metadata_document()
    print(f"Upgraded {doc.stem} to v{meta.container.datadoc.document_version}")  # noqa: T201



import pytest
import requests
from bs4 import BeautifulSoup

from dapla_metadata.datasets.statistic_subject_mapping import PrimarySubject
from dapla_metadata.datasets.statistic_subject_mapping import SecondarySubject
from dapla_metadata.datasets.statistic_subject_mapping import StatisticSubjectMapping
from tests.datasets.constants import TEST_RESOURCES_DIRECTORY


def test_no_source_url(thread_pool_executor):
    subject_mapping = StatisticSubjectMapping(thread_pool_executor, None)
    subject_mapping.wait_for_external_result()
    assert subject_mapping.primary_subjects == []


def test_extract_titles():
    xml_string = '<titler><tittel sprak="no">Partifinansiering</tittel><tittel sprak="en">Funding of political parties</tittel></titler>'
    soup = BeautifulSoup(xml_string, features="xml")
    assert StatisticSubjectMapping._extract_titles(soup) == {  # noqa: SLF001
        "no": "Partifinansiering",
        "en": "Funding of political parties",
    }


STATISTICAL_SUBJECT_STRUCTURE_DIR = "statistical_subject_structure"


@pytest.mark.parametrize(
    ("subject_xml_file_path", "expected"),
    [
        (
            TEST_RESOURCES_DIRECTORY / STATISTICAL_SUBJECT_STRUCTURE_DIR / "simple.xml",
            [
                PrimarySubject(
                    titles={"en": "aa english", "no": "aa norwegian"},
                    subject_code="aa",
                    secondary_subjects=[
                        SecondarySubject(
                            titles={"en": "aa00 english", "no": "aa00 norwegian"},
                            subject_code="aa00",
                            statistic_short_names=["aa_kortnvan"],
                        ),
                    ],
                ),
            ],
        ),
        (
            TEST_RESOURCES_DIRECTORY / STATISTICAL_SUBJECT_STRUCTURE_DIR / "empty.xml",
            [],
        ),
        (
            TEST_RESOURCES_DIRECTORY
            / STATISTICAL_SUBJECT_STRUCTURE_DIR
            / "missing_language.xml",
            [
                PrimarySubject(
                    titles={
                        "en": "aa english",
                    },
                    subject_code="aa",
                    secondary_subjects=[
                        SecondarySubject(
                            titles={
                                "no": "aa00 norwegian",
                                "en": "aa00 english",
                            },
                            subject_code="aa00",
                            statistic_short_names=["aa_kortnvan"],
                        ),
                        SecondarySubject(
                            titles={
                                "no": "aa01 norwegian",
                            },
                            subject_code="aa01",
                            statistic_short_names=["aa_kortnvan_01"],
                        ),
                    ],
                ),
                PrimarySubject(
                    titles={
                        "en": "ab english",
                    },
                    subject_code="ab",
                    secondary_subjects=[
                        SecondarySubject(
                            titles={
                                "no": "ab00 norwegian",
                                "en": "ab00 english",
                            },
                            subject_code="ab00",
                            statistic_short_names=["ab_kortnvan"],
                        ),
                        SecondarySubject(
                            titles={
                                "en": "ab01 english",
                            },
                            subject_code="ab01",
                            statistic_short_names=["ab_kortnvan_01"],
                        ),
                    ],
                ),
            ],
        ),
    ],
)
@pytest.mark.usefixtures("_mock_fetch_statistical_structure")
def test_read_in_statistical_structure(
    subject_mapping_fake_statistical_structure: StatisticSubjectMapping,
    expected: list[PrimarySubject],
) -> None:
    subject_mapping_fake_statistical_structure.wait_for_external_result()
    assert subject_mapping_fake_statistical_structure.primary_subjects == expected


@pytest.mark.parametrize(
    ("statistic_short_name", "expected_secondary_subject"),
    [
        ("ab_kortnvan", "ab00"),
        ("aa_kortnvan", "aa00"),
        ("ab_kortnvan_01", "ab01"),
        ("aa_kortnvan_01", "aa01"),
        ("unknown_name", None),
        (None, None),
    ],
)
@pytest.mark.usefixtures("_mock_fetch_statistical_structure")
def test_get_secondary_subject(
    subject_mapping_fake_statistical_structure: StatisticSubjectMapping,
    statistic_short_name: str,
    expected_secondary_subject: str,
) -> None:
    subject_mapping_fake_statistical_structure.wait_for_external_result()
    assert (
        subject_mapping_fake_statistical_structure.get_secondary_subject(
            statistic_short_name,
        )
        == expected_secondary_subject
    )


@pytest.fixture
def subject_mapping_http_exception(
    requests_mock,
    exception_to_raise,
    thread_pool_executor,
) -> StatisticSubjectMapping:
    requests_mock.get(
        "http://test.some.url.com",
        exc=exception_to_raise,
    )
    return StatisticSubjectMapping(thread_pool_executor, "http://test.some.url.com")


@pytest.mark.parametrize(
    ("exception_to_raise"),
    [
        (requests.exceptions.ConnectTimeout),
        (requests.exceptions.HTTPError),
        (requests.exceptions.ReadTimeout),
        (requests.exceptions.ConnectionError),
    ],
)
def test_subject_mapping_http_exception(
    subject_mapping_http_exception: StatisticSubjectMapping,
) -> None:
    subject_mapping_http_exception.wait_for_external_result()
    assert subject_mapping_http_exception.primary_subjects == []



import functools
from collections.abc import Callable
from http import HTTPStatus
from pathlib import Path
from unittest.mock import MagicMock
from unittest.mock import Mock
from unittest.mock import patch

import pytest
import urllib3

from dapla_metadata.variable_definitions._generated.vardef_client.api.variable_definitions_api import (
    VariableDefinitionsApi,
)
from dapla_metadata.variable_definitions._generated.vardef_client.api_client import (
    ApiClient,
)
from dapla_metadata.variable_definitions._generated.vardef_client.configuration import (
    Configuration,
)
from dapla_metadata.variable_definitions._generated.vardef_client.models.draft import (
    Draft,
)
from dapla_metadata.variable_definitions._generated.vardef_client.models.get_vardok_vardef_mapping_by_id200_response import (
    GetVardokVardefMappingById200Response,
)
from dapla_metadata.variable_definitions._generated.vardef_client.models.vardok_id_response import (
    VardokIdResponse,
)
from dapla_metadata.variable_definitions._generated.vardef_client.models.vardok_vardef_id_pair_response import (
    VardokVardefIdPairResponse,
)
from dapla_metadata.variable_definitions._generated.vardef_client.models.variable_status import (
    VariableStatus,
)
from dapla_metadata.variable_definitions._utils._client import VardefClient
from dapla_metadata.variable_definitions.exceptions import STATUS_EXPLANATIONS
from dapla_metadata.variable_definitions.exceptions import VardefClientError
from dapla_metadata.variable_definitions.exceptions import VariableNotFoundError
from dapla_metadata.variable_definitions.vardef import Vardef
from dapla_metadata.variable_definitions.vardok_id import VardokId
from dapla_metadata.variable_definitions.variable_definition import VariableDefinition
from tests.utils.constants import VARDEF_EXAMPLE_DATE
from tests.utils.constants import VARDEF_EXAMPLE_DEFINITION_ID
from tests.utils.constants import VARDEF_EXAMPLE_INVALID_ID
from tests.utils.constants import VARDEF_EXAMPLE_SHORT_NAME
from tests.variable_definitions.conftest import sample_variable_definition
from tests.variable_definitions.conftest import unknown_variable_definition

PATCH_ID = 2


def test_list_variable_definitions(client_configuration: Configuration):
    VardefClient.set_config(client_configuration)
    assert Vardef.list_variable_definitions() == []


def test_list_variable_definitions_with_date_of_validity(
    client_configuration: Configuration,
):
    VardefClient.set_config(client_configuration)
    assert isinstance(
        Vardef.list_variable_definitions(date_of_validity=VARDEF_EXAMPLE_DATE)[0],
        VariableDefinition,
    )


def test_get_variable_definition_by_id(client_configuration: Configuration):
    VardefClient.set_config(client_configuration)
    landbak = Vardef.get_variable_definition_by_id(
        variable_definition_id=VARDEF_EXAMPLE_DEFINITION_ID,
    )
    assert isinstance(landbak, VariableDefinition)
    assert landbak.classification_reference == "91"


def test_get_variable_definition_by_short_name(client_configuration: Configuration):
    VardefClient.set_config(client_configuration)
    landbak = Vardef.get_variable_definition_by_shortname(
        short_name=VARDEF_EXAMPLE_SHORT_NAME,
    )
    assert isinstance(landbak, VariableDefinition)
    assert landbak.classification_reference == "91"


def test_get_variable_definition_by_nonexistent_short_name(
    client_configuration: Configuration,
):
    VardefClient.set_config(client_configuration)
    short_name = "nonexistent"
    mock_response = []
    with (
        patch.object(
            VariableDefinitionsApi,
            "list_variable_definitions",
            return_value=mock_response,
        ),
        pytest.raises(
            VariableNotFoundError,
            match=f"Variable with short name {short_name} not found",
        ),
    ):
        Vardef.get_variable_definition_by_shortname(short_name=short_name)


def test_get_variable_definition_multiple_variables_returned(
    client_configuration: Configuration,
):
    VardefClient.set_config(client_configuration)
    short_name = "multiple"
    mock_response = ["variable", "variable"]
    with (
        patch.object(
            VariableDefinitionsApi,
            "list_variable_definitions",
            return_value=mock_response,
        ),
        pytest.raises(
            VariableNotFoundError,
            match=f"Lookup by short name {short_name} found multiple variables which should not be possible",
        ),
    ):
        Vardef.get_variable_definition_by_shortname(short_name=short_name)


@patch.object(
    ApiClient,
    "call_api",
)
@pytest.mark.parametrize(
    ("exception"),
    [
        urllib3.exceptions.ProtocolError(),
        urllib3.exceptions.MaxRetryError(
            pool=urllib3.connectionpool.HTTPConnectionPool(host="example.com"),
            url="http://www.example.com",
        ),
    ],
)
def test_urllib_exceptions(
    mock_call_api: Mock,
    exception: urllib3.exceptions.HTTPError,
):
    mock_call_api.side_effect = exception
    with pytest.raises(VardefClientError):
        Vardef.get_variable_definition_by_id(VARDEF_EXAMPLE_DEFINITION_ID)


@pytest.mark.parametrize(
    ("method"),
    [
        functools.partial(
            Vardef.get_variable_definition_by_id,
            variable_definition_id=VARDEF_EXAMPLE_INVALID_ID,
        ),
        unknown_variable_definition().list_validity_periods,
        unknown_variable_definition().list_patches,
        functools.partial(sample_variable_definition().get_patch, patch_id=244),
    ],
)
def test_not_found(
    client_configuration: Configuration,
    method: Callable,
):
    VardefClient.set_config(client_configuration)
    with pytest.raises(VardefClientError) as e:
        method()
    assert e.value.status == HTTPStatus.NOT_FOUND
    assert str(e.value) == (
        STATUS_EXPLANATIONS[HTTPStatus.NOT_FOUND] + "\nDetail: Not found"
    )


def test_create_draft(
    client_configuration: Configuration,
    draft: Draft,
):
    VardefClient.set_config(client_configuration)
    my_draft = Vardef.create_draft(
        draft=draft,
    )
    assert isinstance(my_draft, VariableDefinition)
    assert my_draft.id is not None
    assert my_draft.patch_id == 1
    assert my_draft.variable_status == VariableStatus.DRAFT


@pytest.mark.usefixtures("work_folder_complete_patch_output")
def test_create_draft_from_file():
    my_draft = Vardef.create_draft_from_file()
    assert isinstance(my_draft, VariableDefinition)
    assert my_draft.id is not None
    assert my_draft.patch_id == 1
    assert my_draft.variable_status == VariableStatus.DRAFT


def test_create_draft_from_file_specify_path(work_folder_complete_patch_output: Path):
    my_draft = Vardef.create_draft_from_file(work_folder_complete_patch_output)
    assert isinstance(my_draft, VariableDefinition)
    assert my_draft.id is not None
    assert my_draft.patch_id == 1
    assert my_draft.variable_status == VariableStatus.DRAFT


def test_migrate_from_vardok(
    client_configuration: Configuration,
):
    VardefClient.set_config(client_configuration)
    my_draft = Vardef.migrate_from_vardok(
        vardok_id="1607",
    )
    assert isinstance(my_draft, VariableDefinition)
    assert my_draft.id is not None
    assert my_draft.patch_id == 1
    assert my_draft.variable_status == VariableStatus.DRAFT


def test_short_name_exist():
    mock_variable = MagicMock()
    mock_variable.short_name = "test_name"

    mock_variable1 = MagicMock()
    mock_variable1.short_name = "random_name"

    mock_variable2 = MagicMock()
    mock_variable2.short_name = "pluu_h"

    mock_variable3 = MagicMock()
    mock_variable3.short_name = "another_name"

    with patch(
        "dapla_metadata.variable_definitions.vardef.Vardef.list_variable_definitions",
        return_value=[mock_variable, mock_variable1, mock_variable2, mock_variable3],
    ):
        result = Vardef.does_short_name_exist("test_name")
        assert result is True


def test_short_name_does_not_exist():
    mock_variable = MagicMock()
    mock_variable.short_name = "test_name"

    mock_variable1 = MagicMock()
    mock_variable1.short_name = "random_name"
    with patch(
        "dapla_metadata.variable_definitions.vardef.Vardef.list_variable_definitions",
        return_value=[mock_variable, mock_variable1],
    ):
        result = Vardef.does_short_name_exist("org_name")
        assert result is False


def test_short_name_with_whitespace():
    mock_variable = MagicMock()
    mock_variable.short_name = "test_name"

    mock_variable1 = MagicMock()
    mock_variable1.short_name = "random_name"
    with patch(
        "dapla_metadata.variable_definitions.vardef.Vardef.list_variable_definitions",
        return_value=[mock_variable, mock_variable1],
    ):
        result = Vardef.does_short_name_exist("test_name  ")
        assert result is True


def test_list_vardok_vardef_mappings(
    client_configuration: Configuration,
):
    VardefClient.set_config(client_configuration)
    vardok_vardef_mapping = Vardef.list_vardok_vardef_mapping()[0]
    assert isinstance(
        vardok_vardef_mapping,
        VardokVardefIdPairResponse,
    )
    assert vardok_vardef_mapping.vardef_id == "A1699WWE"
    assert vardok_vardef_mapping.vardok_id == "25"


def test_empty_list_vardok_vardef_mappings(
    client_configuration: Configuration,
):
    VardefClient.set_config(client_configuration)
    with patch(
        "dapla_metadata.variable_definitions._generated.vardef_client.api.data_migration_api.DataMigrationApi.get_vardok_vardef_mapping",
        return_value=[],
    ):
        result = Vardef.list_vardok_vardef_mapping()
        assert len(result) == 0


def test_get_variable_by_vardok_id(
    client_configuration: Configuration,
):
    VardefClient.set_config(client_configuration)
    vardok_vardef_mapping = Vardef.get_variable_definition_by_vardok_id("1607")

    assert isinstance(vardok_vardef_mapping, VariableDefinition)
    assert vardok_vardef_mapping.short_name == "landbak"


def test_get_variable_by_vardok_id_but_supplying_vardef_id(
    client_configuration: Configuration,
):
    VardefClient.set_config(client_configuration)

    response_data = VardokIdResponse(vardok_id="wypvb3wd")
    wrapped_response = GetVardokVardefMappingById200Response(
        actual_instance=response_data
    )
    with (
        patch(
            "dapla_metadata.variable_definitions._generated.vardef_client.api.data_migration_api.DataMigrationApi.get_vardok_vardef_mapping_by_id",
            return_value=wrapped_response,
        ),
        pytest.raises(TypeError),
    ):
        Vardef.get_variable_definition_by_vardok_id("wypvb3wd")


def test_get_vardok_id_by_short_name(client_configuration: Configuration):
    VardefClient.set_config(client_configuration)

    response_data = VardokIdResponse(vardok_id="1607")
    wrapped_response = GetVardokVardefMappingById200Response(
        actual_instance=response_data
    )
    with patch(
        "dapla_metadata.variable_definitions._generated.vardef_client.api.data_migration_api.DataMigrationApi.get_vardok_vardef_mapping_by_id",
        return_value=wrapped_response,
    ):
        vardok_vardef_mapping = Vardef.get_vardok_id_by_short_name("landbak")

        assert isinstance(vardok_vardef_mapping, VardokId)
        assert vardok_vardef_mapping.vardok_id == "1607"



from dapla_metadata.variable_definitions._generated.vardef_client.api_client import (
    ApiClient,
)
from dapla_metadata.variable_definitions._generated.vardef_client.configuration import (
    Configuration,
)
from dapla_metadata.variable_definitions._utils.config import (
    get_vardef_client_configuration,
)
from dapla_metadata.variable_definitions._utils.config import refresh_access_token


class VardefClient:
    """Configure an ApiClient object and make it availabe to client code."""

    _client: ApiClient | None = None
    _config: Configuration | None = None

    @classmethod
    def get_config(cls) -> Configuration | None:
        """Get the client configuration object."""
        return cls._config

    @classmethod
    def set_config(cls, config: Configuration) -> None:
        """Set the client configuration object."""
        cls._config = config

    @classmethod
    def get_client(cls) -> ApiClient:
        """Configure and return an ApiClient for use with the Vardef API.

        Retrieves a fresh access token for each request such that we
        don't attempt to use an expired token.

        Returns:
            ApiClient: The configured client.
        """
        if not cls._client:
            if not cls._config:
                cls._config = get_vardef_client_configuration()
            cls._client = ApiClient(cls._config)
        cls._client.configuration.access_token = refresh_access_token()
        return cls._client
